from cognite.client.data_classes._base import *


class ServiceAccount(CogniteResource):
    """No description.

    Args:
        name (str): Unique name of the service account
        groups (List[int]): List of group ids
        id (int): No description.
        is_deleted (bool): If this service account has been logically deleted
        deleted_time (int): Time of deletion
        cognite_client (CogniteClient): The client to associate with this object.
    """

    def __init__(
        self,
        name: str = None,
        groups: List[int] = None,
        id: int = None,
        is_deleted: bool = None,
        deleted_time: int = None,
        cognite_client=None,
    ):
        self.name = name
        self.groups = groups
        self.id = id
        self.is_deleted = is_deleted
        self.deleted_time = deleted_time
        self._cognite_client = cognite_client


class ServiceAccountList(CogniteResourceList):
    _RESOURCE = ServiceAccount
    _ASSERT_CLASSES = False


class APIKey(CogniteResource):
    """No description.

    Args:
        id (int): The internal ID for the API key.
        service_account_id (int): The ID of the service account.
        created_time (int): The time of creation in Unix milliseconds.
        status (str): The status of the API key.
        value (str): The API key to be used against the API.
        cognite_client (CogniteClient): The client to associate with this object.
    """

    def __init__(
        self,
        id: int = None,
        service_account_id: int = None,
        created_time: int = None,
        status: str = None,
        value: str = None,
        cognite_client=None,
    ):
        self.id = id
        self.service_account_id = service_account_id
        self.created_time = created_time
        self.status = status
        self.value = value
        self._cognite_client = cognite_client


class APIKeyList(CogniteResourceList):
    _RESOURCE = APIKey
    _ASSERT_CLASSES = False


class Group(CogniteResource):
    """No description.

    Args:
        name (str): Name of the group
        source_id (str): ID of the group in the source. If this is the same ID as a group in the IDP, a service account in that group will implicitly be a part of this group as well.
        capabilities (List[Dict[str, Any]]): No description.
        id (int): No description.
        is_deleted (bool): No description.
        deleted_time (int): No description.
        cognite_client (CogniteClient): The client to associate with this object.
    """

    def __init__(
        self,
        name: str = None,
        source_id: str = None,
        capabilities: List[Dict[str, Any]] = None,
        id: int = None,
        is_deleted: bool = None,
        deleted_time: int = None,
        cognite_client=None,
    ):
        self.name = name
        self.source_id = source_id
        self.capabilities = capabilities
        self.id = id
        self.is_deleted = is_deleted
        self.deleted_time = deleted_time
        self._cognite_client = cognite_client


class GroupList(CogniteResourceList):
    _RESOURCE = Group
    _ASSERT_CLASSES = False


class SecurityCategory(CogniteResource):
    """No description.

    Args:
        name (str): Name of the security category
        id (int): Id of the security category
        cognite_client (CogniteClient): The client to associate with this object.
    """

    def __init__(self, name: str = None, id: int = None, cognite_client=None):
        self.name = name
        self.id = id
        self._cognite_client = cognite_client


class SecurityCategoryList(CogniteResourceList):
    _RESOURCE = SecurityCategory
    _ASSERT_CLASSES = False


class ProjectSpec(CogniteResponse):
    """A cdf project spec
    Args:
        url_name (str): The url name for the project
        groups (List[int]): Group ids in the project
    """

    def __init__(self, url_name: str, groups: List[int]) -> None:
        self.url_name = url_name
        self.groups = groups

    @classmethod
    def _load(cls, api_response):
        return cls(url_name=api_response["projectUrlName"], groups=api_response["groups"])


class TokenInspection(CogniteResponse):
    """Current login status

    Args:
        subject (str): Subject (sub claim) of JWT.
        projects (List[ProjectSpec]): Projects this token is valid for.
        capabilities (List[Dict]): Capabilities associated with this token.
    """

    def __init__(self, subject: str, projects: List[ProjectSpec], capabilities: List[Dict]):
        self.subject = subject
        self.projects = projects
        self.capabilities = capabilities

    @classmethod
    def _load(cls, api_response):
        return cls(
            subject=api_response["subject"],
            projects=[ProjectSpec._load(p) for p in api_response["projects"]],
            capabilities=api_response["capabilities"],
        )

    def dump(self, camel_case: bool = False) -> Dict[str, Any]:
        dumped = {
            "subject": self.subject,
            "projects": [p.dump(camel_case=camel_case) for p in self.projects],
            "capabilities": self.capabilities,
        }
        if camel_case:
            dumped = {utils._auxiliary.to_camel_case(key): value for key, value in dumped.items()}
        return dumped



from datetime import datetime
from unittest import mock

import pytest

import cognite.client.utils._time
from cognite.client import utils
from cognite.client.data_classes import EndTimeFilter, Event, EventFilter, EventUpdate
from cognite.client.exceptions import CogniteNotFoundError
from tests.utils import set_request_limit


@pytest.fixture
def new_event(cognite_client):
    event = cognite_client.events.create(Event(type="test__py__sdk"))
    yield event
    cognite_client.events.delete(id=event.id)
    assert cognite_client.events.retrieve(event.id) is None


@pytest.fixture
def post_spy(cognite_client):
    with mock.patch.object(cognite_client.events, "_post", wraps=cognite_client.events._post) as _:
        yield


@pytest.fixture
def root_test_asset(cognite_client):
    for asset in cognite_client.assets(root=True):
        if asset.name.startswith("test__"):
            return asset


class TestEventsAPI:
    def test_retrieve(self, cognite_client):
        res = cognite_client.events.list(limit=1)
        assert res[0] == cognite_client.events.retrieve(res[0].id)

    def test_retrieve_multiple(self, cognite_client, root_test_asset):
        res_listed_ids = [
            e.id for e in cognite_client.events.list(limit=2, root_asset_ids=[{"id": root_test_asset.id}])
        ]
        res_lookup_ids = [e.id for e in cognite_client.events.retrieve_multiple(res_listed_ids)]
        for listed_id in res_listed_ids:
            assert listed_id in res_lookup_ids

    def test_retrieve_unknown(self, cognite_client):
        res = cognite_client.events.list(limit=1)
        with pytest.raises(CogniteNotFoundError):
            cognite_client.events.retrieve_multiple(ids=[res[0].id], external_ids=["this does not exist"])
        retr = cognite_client.events.retrieve_multiple(
            ids=[res[0].id], external_ids=["this does not exist"], ignore_unknown_ids=True
        )
        assert 1 == len(retr)

    def test_list(self, cognite_client, post_spy):
        with set_request_limit(cognite_client.events, 10):
            res = cognite_client.events.list(limit=20)

        assert 20 == len(res)
        assert 2 == cognite_client.events._post.call_count

    def test_list_ongoing(self, cognite_client):
        res = cognite_client.events.list(end_time=EndTimeFilter(is_null=True), limit=10)

        assert len(res) > 0

    def test_aggregation(self, cognite_client, new_event):
        res_aggregate = cognite_client.events.aggregate(filter=EventFilter(type="test__py__sdk"))
        assert res_aggregate[0].count > 0

    def test_partitioned_list(self, cognite_client):
        # stop race conditions by cutting off max created time
        maxtime = utils.timestamp_to_ms(datetime(2019, 5, 25, 17, 30))
        res_flat = cognite_client.events.list(limit=None, type="test-data-populator", start_time={"max": maxtime})
        res_part = cognite_client.events.list(
            partitions=8, type="test-data-populator", start_time={"max": maxtime}, limit=None
        )
        assert len(res_flat) > 0
        assert len(res_flat) == len(res_part)
        assert {a.id for a in res_flat} == {a.id for a in res_part}

    def test_compare_partitioned_gen_and_list(self, cognite_client, post_spy):
        # stop race conditions by cutting off max created time
        maxtime = utils.timestamp_to_ms(datetime(2019, 5, 25, 17, 30))
        res_generator = cognite_client.events(partitions=8, limit=None, created_time={"max": maxtime})
        res_list = cognite_client.events.list(partitions=8, limit=None, created_time={"max": maxtime})
        assert {a.id for a in res_generator} == {a.id for a in res_list}

    def test_assetid_list(self, cognite_client):
        res = cognite_client.events.list(
            limit=None, type="test-data-populator", asset_external_ids=["a", "b"], asset_ids=[0, 1]
        )
        assert 0 == len(res)

    def test_search(self, cognite_client):
        res = cognite_client.events.search(
            filter=EventFilter(start_time={"min": cognite.client.utils._time.timestamp_to_ms("2d-ago")})
        )
        assert len(res) > 0

    def test_update(self, cognite_client, new_event):
        update_event = EventUpdate(new_event.id).metadata.set({"bla": "bla"})
        res = cognite_client.events.update(update_event)
        assert {"bla": "bla"} == res.metadata
        update_event2 = EventUpdate(new_event.id).metadata.set(None)
        res2 = cognite_client.events.update(update_event2)
        assert res2.metadata == {}

    def test_delete_with_nonexisting(self, cognite_client):
        a = cognite_client.events.create(Event())
        cognite_client.events.delete(id=a.id, external_id="this event does not exist", ignore_unknown_ids=True)
        assert cognite_client.events.retrieve(id=a.id) is None



import time
import uuid
from tempfile import NamedTemporaryFile, TemporaryDirectory

import pytest

from cognite.client.data_classes import (
    FileMetadata,
    FileMetadataFilter,
    FileMetadataUpdate,
    GeoLocation,
    GeoLocationFilter,
    Geometry,
    GeometryFilter,
    Label,
    LabelDefinition,
)
from cognite.client.utils._auxiliary import random_string


@pytest.fixture(scope="class")
def mock_geo_location():
    geometry = Geometry(type="LineString", coordinates=[[30, 10], [10, 30], [40, 40]])
    yield GeoLocation(type="Feature", geometry=geometry, properties=dict())


@pytest.fixture(scope="class")
def new_file(cognite_client):
    res = cognite_client.files.upload_bytes(content="blabla", name="myspecialfile")
    while True:
        if cognite_client.files.retrieve(id=res.id).uploaded:
            break
        time.sleep(0.5)
    yield res
    cognite_client.files.delete(id=res.id)
    assert cognite_client.files.retrieve(id=res.id) is None


@pytest.fixture(scope="class")
def empty_file(cognite_client):
    name = "empty_" + random_string(10)
    res = cognite_client.files.upload_bytes(content=b"", name=name, external_id=name)
    while True:
        if cognite_client.files.retrieve(id=res.id).uploaded:
            break
        time.sleep(0.5)
    yield res
    cognite_client.files.delete(id=res.id)


@pytest.fixture(scope="class")
def new_file_with_geoLocation(mock_geo_location, cognite_client):
    res = cognite_client.files.upload_bytes(content="blabla", name="geoLocationFile", geo_location=mock_geo_location)
    while True:
        if cognite_client.files.retrieve(id=res.id).uploaded:
            break
        time.sleep(0.5)
    yield res
    cognite_client.files.delete(id=res.id)
    assert cognite_client.files.retrieve(id=res.id) is None


@pytest.fixture(scope="class")
def new_file_with_label(cognite_client):
    label_external_id = uuid.uuid4().hex[0:20]
    label = cognite_client.labels.create(LabelDefinition(external_id=label_external_id, name="mandatory"))
    file = cognite_client.files.upload_bytes(
        content="blabla", name="myspecialfile", labels=[Label(external_id=label_external_id)]
    )
    while True:
        if cognite_client.files.retrieve(id=file.id).uploaded:
            break
        time.sleep(0.5)
    yield file, label.external_id
    cognite_client.files.delete(id=file.id)
    cognite_client.labels.delete(external_id=label_external_id)
    assert cognite_client.files.retrieve(id=file.id) is None


@pytest.fixture(scope="class")
def test_files(cognite_client):
    files = {}
    for file in cognite_client.files:
        if file.name in ["a.txt", "b.txt", "c.txt", "big.txt"]:
            files[file.name] = file
    return files


A_WHILE_AGO = {"max": int(time.time() - 1800) * 1000}


class TestFilesAPI:
    def test_create(self, cognite_client):
        file_metadata = FileMetadata(name="mytestfile")
        returned_file_metadata, upload_url = cognite_client.files.create(file_metadata)
        assert returned_file_metadata.uploaded is False
        cognite_client.files.delete(id=returned_file_metadata.id)

    def test_create_with_geoLocation(self, cognite_client, mock_geo_location):
        file_metadata = FileMetadata(name="mytestfile", geo_location=mock_geo_location)
        returned_file_metadata, upload_url = cognite_client.files.create(file_metadata)
        assert returned_file_metadata.uploaded is False
        assert returned_file_metadata.geo_location == mock_geo_location
        cognite_client.files.delete(id=returned_file_metadata.id)

    def test_retrieve(self, cognite_client):
        res = cognite_client.files.list(name="big.txt", limit=1)
        assert res[0] == cognite_client.files.retrieve(res[0].id)

    def test_retrieve_multiple(self, cognite_client):
        res = cognite_client.files.list(uploaded_time=A_WHILE_AGO, limit=2)
        assert res == cognite_client.files.retrieve_multiple([f.id for f in res])

    def test_retrieve_download_urls(self, cognite_client):
        f1 = cognite_client.files.upload_bytes(b"f1", external_id=random_string(10), name="bla")
        f2 = cognite_client.files.upload_bytes(b"f2", external_id=random_string(10), name="bla")
        download_links = cognite_client.files.retrieve_download_urls(id=f1.id, external_id=f2.external_id)
        assert len(download_links.values()) == 2
        assert download_links[f1.id].startswith("http")
        assert download_links[f2.external_id].startswith("http")

    def test_list(self, cognite_client):
        res = cognite_client.files.list(limit=4)
        assert 4 == len(res)

    def test_aggregate(self, cognite_client):
        res = cognite_client.files.aggregate(filter=FileMetadataFilter(name="big.txt"))
        assert res[0].count > 0

    def test_search(self, cognite_client):
        res = cognite_client.files.search(name="big.txt", filter=FileMetadataFilter(created_time={"min": 0}))
        assert len(res) > 0

    def test_update(self, cognite_client, new_file):
        update_file = FileMetadataUpdate(new_file.id).metadata.set({"bla": "bla"})
        res = cognite_client.files.update(update_file)
        assert {"bla": "bla"} == res.metadata

    def test_update_directory(self, cognite_client, new_file):
        dir = "/some/directory"
        res = cognite_client.files.update(FileMetadata(id=new_file.id, directory=dir))
        assert res.directory == dir

    def test_download(self, cognite_client, test_files):
        test_file = test_files["a.txt"]
        res = cognite_client.files.download_bytes(id=test_file.id)
        assert b"a" == res

    def test_download_new_file(self, cognite_client, new_file):
        assert b"blabla" == cognite_client.files.download_bytes(id=new_file.id)

    def test_download_empty_file(self, cognite_client, empty_file):
        content = cognite_client.files.download_bytes(external_id=empty_file.external_id)
        assert content == b""

        with TemporaryDirectory() as tmpdir:
            cognite_client.files.download(directory=tmpdir, external_id=empty_file.external_id)

        with NamedTemporaryFile() as tmpfile:
            cognite_client.files.download_to_path(path=tmpfile.name, external_id=empty_file.external_id)

    def test_retrieve_file_with_labels(self, cognite_client, new_file_with_label):
        file, label_external_id = new_file_with_label
        res = cognite_client.files.retrieve(id=file.id)
        assert len(res.labels) == 1
        assert res.labels[0].external_id == label_external_id

    def test_filter_file_on_geoLocation(self, cognite_client, new_file_with_geoLocation, mock_geo_location):
        max_retries = 10
        geometry_filter = GeometryFilter(type="Point", coordinates=[30, 10])
        geo_location_filter = GeoLocationFilter(relation="intersects", shape=geometry_filter)
        res = cognite_client.files.list(geo_location=geo_location_filter)
        for _ in range(max_retries):
            if len(res) > 0:
                break
            time.sleep(0.2)
            res = cognite_client.files.list(geo_location=geo_location_filter)
        assert res[0].geo_location == new_file_with_geoLocation.geo_location



import json

from cognite.client.data_classes import AggregateResultItem, TimestampRange


class TestGenerated:
    def test_timestamprange(self):
        tsr = TimestampRange(min=1, max=2)
        assert 1 == tsr.min
        assert 2 == tsr.max
        assert 1 == tsr["min"]
        assert 2 == tsr["max"]
        tsr.max = 3
        assert 3 == tsr["max"]
        tsr["max"] = 4
        assert 4 == tsr["max"]
        assert {"max": 4, "min": 1} == json.loads(json.dumps(tsr))

    def test_one_missing(self):
        tsr = TimestampRange(min=1)
        assert {"min": 1} == json.loads(json.dumps(tsr))

    def test_typo(self):
        tsr = TimestampRange()
        assert tsr.min is None
        assert tsr.get("min") is None
        assert {} == json.loads(json.dumps(tsr))
        tsr["max"] = 4
        assert {"max": 4} == json.loads(json.dumps(tsr))
        tsr.mun = 4
        assert {"max": 4} == json.loads(json.dumps(tsr))

    def test_camels(self):
        ag = AggregateResultItem(child_count=23)
        assert 23 == ag.child_count
        assert 23 == ag["childCount"]
        assert ag.get("child_count") is None
        assert {"childCount": 23} == json.loads(json.dumps(ag))



import math
import os
from collections import namedtuple

import pytest
from requests import Response

from cognite.client import CogniteClient
from cognite.client._api_client import APIClient
from cognite.client.data_classes._base import *
from cognite.client.exceptions import CogniteAPIError, CogniteNotFoundError
from cognite.client.utils._client_config import ClientConfig
from tests.utils import jsgz_load, set_env_var, set_request_limit

BASE_URL = "http://localtest.com/api/1.0/projects/test-project"
URL_PATH = "/someurl"

RESPONSE = {"any": "ok"}


@pytest.fixture(scope="module")
def api_client_with_api_key(cognite_client):
    return APIClient(
        ClientConfig(
            project="test-project",
            api_key="abc",
            base_url=BASE_URL,
            max_workers=1,
            headers={"x-cdp-app": "python-sdk-integration-tests"},
        ),
        cognite_client=cognite_client,
    )


@pytest.fixture(scope="module")
def api_client_with_token_factory(cognite_client):
    return APIClient(
        ClientConfig(
            project="test-project",
            base_url=BASE_URL,
            max_workers=1,
            headers={"x-cdp-app": "python-sdk-integration-tests"},
            token=lambda: "abc",
        ),
        cognite_client=cognite_client,
    )


@pytest.fixture(scope="module")
def api_client_with_token(cognite_client):
    return APIClient(
        ClientConfig(
            project="test-project",
            base_url=BASE_URL,
            max_workers=1,
            headers={"x-cdp-app": "python-sdk-integration-tests"},
            token="abc",
        ),
        cognite_client=cognite_client,
    )


RequestCase = namedtuple("RequestCase", ["name", "method", "kwargs"])


class TestBasicRequests:
    @pytest.fixture
    def mock_all_requests_ok(self, rsps):
        rsps.assert_all_requests_are_fired = False
        for method in [rsps.GET, rsps.PUT, rsps.POST, rsps.DELETE]:
            rsps.add(method, BASE_URL + URL_PATH, status=200, json=RESPONSE)
        yield rsps

    @pytest.fixture
    def mock_all_requests_fail(self, rsps):
        rsps.assert_all_requests_are_fired = False
        for method in [rsps.GET, rsps.PUT, rsps.POST, rsps.DELETE]:
            rsps.add(method, BASE_URL + URL_PATH, status=400, json={"error": "Client error"})
            rsps.add(method, BASE_URL + URL_PATH, status=500, body="Server error")
            rsps.add(method, BASE_URL + URL_PATH, status=500, json={"error": "Server error"})
            rsps.add(method, BASE_URL + URL_PATH, status=400, json={"error": {"code": 400, "message": "Client error"}})

    request_cases = [
        lambda api_client: RequestCase(
            name="post", method=api_client._post, kwargs={"url_path": URL_PATH, "json": {"any": "ok"}}
        ),
        lambda api_client: RequestCase(name="get", method=api_client._get, kwargs={"url_path": URL_PATH}),
        lambda api_client: RequestCase(name="delete", method=api_client._delete, kwargs={"url_path": URL_PATH}),
        lambda api_client: RequestCase(
            name="put", method=api_client._put, kwargs={"url_path": URL_PATH, "json": {"any": "ok"}}
        ),
    ]

    @pytest.mark.parametrize("fn", request_cases)
    def test_requests_ok(self, fn, mock_all_requests_ok, api_client_with_api_key):
        name, method, kwargs = fn(api_client_with_api_key)
        response = method(**kwargs)
        assert response.status_code == 200
        assert response.json() == RESPONSE

        request_headers = mock_all_requests_ok.calls[0].request.headers
        assert "application/json" == request_headers["content-type"]
        assert "application/json" == request_headers["accept"]
        assert api_client_with_api_key._config.api_key == request_headers["api-key"]
        assert "python-sdk-integration-tests" == request_headers["x-cdp-app"]
        assert "User-Agent" in request_headers

    @pytest.mark.usefixtures("mock_all_requests_fail")
    @pytest.mark.parametrize("fn", request_cases)
    def test_requests_fail(self, fn, api_client_with_api_key):
        name, method, kwargs = fn(api_client_with_api_key)
        with pytest.raises(CogniteAPIError, match="Client error") as e:
            method(**kwargs)
        assert e.value.code == 400

        with pytest.raises(CogniteAPIError, match="Server error") as e:
            method(**kwargs)
        assert e.value.code == 500

        with pytest.raises(CogniteAPIError, match="Server error") as e:
            method(**kwargs)
        assert e.value.code == 500

        with pytest.raises(CogniteAPIError, match="Client error | code: 400 | X-Request-ID:") as e:
            method(**kwargs)
        assert e.value.code == 400
        assert e.value.message == "Client error"

    @pytest.mark.usefixtures("disable_gzip")
    def test_request_gzip_disabled(self, rsps, api_client_with_api_key):
        def check_gzip_disabled(request):
            assert "Content-Encoding" not in request.headers
            assert {"any": "OK"} == json.loads(request.body)
            return 200, {}, json.dumps(RESPONSE)

        for method in [rsps.PUT, rsps.POST]:
            rsps.add_callback(method, BASE_URL + URL_PATH, check_gzip_disabled)

        api_client_with_api_key._post(URL_PATH, {"any": "OK"}, headers={})
        api_client_with_api_key._put(URL_PATH, {"any": "OK"}, headers={})

    def test_request_gzip_enabled(self, rsps, api_client_with_api_key):
        def check_gzip_enabled(request):
            assert "Content-Encoding" in request.headers
            assert {"any": "OK"} == jsgz_load(request.body)
            return 200, {}, json.dumps(RESPONSE)

        for method in [rsps.PUT, rsps.POST]:
            rsps.add_callback(method, BASE_URL + URL_PATH, check_gzip_enabled)

        api_client_with_api_key._post(URL_PATH, {"any": "OK"}, headers={})
        api_client_with_api_key._put(URL_PATH, {"any": "OK"}, headers={})

    def test_headers_correct(self, mock_all_requests_ok, api_client_with_api_key):
        api_client_with_api_key._post(URL_PATH, {"any": "OK"}, headers={"additional": "stuff"})
        headers = mock_all_requests_ok.calls[0].request.headers

        assert "gzip, deflate" == headers["accept-encoding"]
        assert "gzip" == headers["content-encoding"]
        assert "CognitePythonSDK:{}".format(utils._auxiliary.get_current_sdk_version()) == headers["x-cdp-sdk"]
        assert "abc" == headers["api-key"]
        assert "stuff" == headers["additional"]

    def test_headers_correct_with_token_factory(self, mock_all_requests_ok, api_client_with_token_factory):
        api_client_with_token_factory._post(URL_PATH, {"any": "OK"})
        headers = mock_all_requests_ok.calls[0].request.headers

        assert "api-key" not in headers
        assert "Bearer {}".format(api_client_with_token_factory._config.token()) == headers["Authorization"]

    def test_headers_correct_with_token(self, mock_all_requests_ok, api_client_with_token):
        api_client_with_token._post(URL_PATH, {"any": "OK"})
        headers = mock_all_requests_ok.calls[0].request.headers

        assert "api-key" not in headers
        assert "Bearer {}".format(api_client_with_token._config.token) == headers["Authorization"]

    @pytest.mark.parametrize("payload", [math.nan, math.inf, -math.inf, {"foo": {"bar": {"baz": [[[math.nan]]]}}}])
    def test__do_request_raises_more_verbose_exception(self, api_client_with_token, payload):
        with pytest.raises(ValueError, match=r"contain NaN\(s\) or \+/\- Inf\!"):
            api_client_with_token._do_request("POST", URL_PATH, json=payload)

    def test__do_request_raises_unmodified_exception(self, api_client_with_token):
        # Create circular ref in payload to raise an arbitrary ValueError
        # we want to make sure we _don't_ modify:
        payload = []
        payload.append(payload)
        with pytest.raises(ValueError) as exc_info:
            api_client_with_token._do_request("POST", URL_PATH, json=payload)
        exc_msg = exc_info.value.args[0]
        assert "contain NaN(s) or +/- Inf!" not in exc_msg


class SomeUpdate(CogniteUpdate):
    @property
    def y(self):
        return PrimitiveUpdate(self, "y")

    @property
    def external_id(self):
        return PrimitiveUpdate(self, "externalId")


class PrimitiveUpdate(CognitePrimitiveUpdate):
    def set(self, value: Any) -> SomeUpdate:
        return self._set(value)


class SomeResource(CogniteResource):
    def __init__(self, x=None, y=None, id=None, external_id=None, cognite_client=None):
        self.x = x
        self.y = y
        self.id = id
        self.external_id = external_id


class SomeResourceList(CogniteResourceList):
    _RESOURCE = SomeResource
    _UPDATE = SomeUpdate


class SomeFilter(CogniteFilter):
    def __init__(self, var_x, var_y):
        self.var_x = var_x
        self.var_y = var_y


class SomeAggregation(dict):
    def __init__(self, count):
        self.count = count


class TestStandardRetrieve:
    def test_standard_retrieve_OK(self, api_client_with_api_key, rsps):
        rsps.add(rsps.GET, BASE_URL + URL_PATH + "/1", status=200, json={"x": 1, "y": 2})
        assert SomeResource(1, 2) == api_client_with_api_key._retrieve(cls=SomeResource, resource_path=URL_PATH, id=1)

    def test_standard_retrieve_not_found(self, api_client_with_api_key, rsps):
        rsps.add(rsps.GET, BASE_URL + URL_PATH + "/1", status=404, json={"error": {"message": "Not Found."}})
        assert api_client_with_api_key._retrieve(cls=SomeResource, resource_path=URL_PATH, id=1) is None

    def test_standard_retrieve_fail(self, api_client_with_api_key, rsps):
        rsps.add(rsps.GET, BASE_URL + URL_PATH + "/1", status=400, json={"error": {"message": "Client Error"}})
        with pytest.raises(CogniteAPIError, match="Client Error") as e:
            api_client_with_api_key._retrieve(cls=SomeResource, resource_path=URL_PATH, id=1)
        assert "Client Error" == e.value.message
        assert 400 == e.value.code

    def test_cognite_client_is_set(self, cognite_client, api_client_with_api_key, rsps):
        rsps.add(rsps.GET, BASE_URL + URL_PATH + "/1", status=200, json={"x": 1, "y": 2})
        assert (
            cognite_client
            == api_client_with_api_key._retrieve(cls=SomeResource, resource_path=URL_PATH, id=1)._cognite_client
        )


class TestStandardRetrieveMultiple:
    @pytest.fixture
    def mock_by_ids(self, rsps):
        rsps.add(rsps.POST, BASE_URL + URL_PATH + "/byids", status=200, json={"items": [{"x": 1, "y": 2}, {"x": 1}]})
        yield rsps

    def test_by_id_no_wrap_OK(self, mock_by_ids, api_client_with_api_key):
        assert SomeResourceList([SomeResource(1, 2), SomeResource(1)]) == api_client_with_api_key._retrieve_multiple(
            cls=SomeResourceList, resource_path=URL_PATH, wrap_ids=False, ids=[1, 2]
        )
        assert {"items": [1, 2]} == jsgz_load(mock_by_ids.calls[0].request.body)

    def test_by_single_id_no_wrap_OK(self, api_client_with_api_key, mock_by_ids):
        assert SomeResource(1, 2) == api_client_with_api_key._retrieve_multiple(
            cls=SomeResourceList, resource_path=URL_PATH, wrap_ids=False, ids=1
        )
        assert {"items": [1]} == jsgz_load(mock_by_ids.calls[0].request.body)

    def test_by_id_wrap_OK(self, api_client_with_api_key, mock_by_ids):
        assert SomeResourceList([SomeResource(1, 2), SomeResource(1)]) == api_client_with_api_key._retrieve_multiple(
            cls=SomeResourceList, resource_path=URL_PATH, wrap_ids=True, ids=[1, 2]
        )
        assert {"items": [{"id": 1}, {"id": 2}]} == jsgz_load(mock_by_ids.calls[0].request.body)

    def test_by_single_id_wrap_OK(self, api_client_with_api_key, mock_by_ids):
        assert SomeResource(1, 2) == api_client_with_api_key._retrieve_multiple(
            cls=SomeResourceList, resource_path=URL_PATH, wrap_ids=True, ids=1
        )
        assert {"items": [{"id": 1}]} == jsgz_load(mock_by_ids.calls[0].request.body)

    def test_by_external_id_wrap_OK(self, api_client_with_api_key, mock_by_ids):
        assert SomeResourceList([SomeResource(1, 2), SomeResource(1)]) == api_client_with_api_key._retrieve_multiple(
            cls=SomeResourceList, resource_path=URL_PATH, wrap_ids=True, external_ids=["1", "2"]
        )
        assert {"items": [{"externalId": "1"}, {"externalId": "2"}]} == jsgz_load(mock_by_ids.calls[0].request.body)

    def test_by_single_external_id_wrap_OK(self, api_client_with_api_key, mock_by_ids):
        assert SomeResource(1, 2) == api_client_with_api_key._retrieve_multiple(
            cls=SomeResourceList, resource_path=URL_PATH, wrap_ids=True, external_ids="1"
        )
        assert {"items": [{"externalId": "1"}]} == jsgz_load(mock_by_ids.calls[0].request.body)

    def test_by_external_id_no_wrap(self, api_client_with_api_key):
        with pytest.raises(ValueError, match="must be wrapped"):
            api_client_with_api_key._retrieve_multiple(
                cls=SomeResourceList, resource_path=URL_PATH, wrap_ids=False, external_ids=["1", "2"]
            )

    def test_retrieve_multiple_ignore_unknown(self, api_client_with_api_key, mock_by_ids):
        assert SomeResourceList([SomeResource(1, 2), SomeResource(1)]) == api_client_with_api_key._retrieve_multiple(
            cls=SomeResourceList,
            resource_path=URL_PATH,
            wrap_ids=True,
            ids=1,
            external_ids=["2"],
            ignore_unknown_ids=True,
        )
        assert {"items": [{"id": 1}, {"externalId": "2"}], "ignoreUnknownIds": True} == jsgz_load(
            mock_by_ids.calls[0].request.body
        )

    def test_id_and_external_id_mixed(self, api_client_with_api_key, mock_by_ids):
        assert SomeResourceList([SomeResource(1, 2), SomeResource(1)]) == api_client_with_api_key._retrieve_multiple(
            cls=SomeResourceList, resource_path=URL_PATH, wrap_ids=True, ids=1, external_ids=["2"]
        )
        assert {"items": [{"id": 1}, {"externalId": "2"}]} == jsgz_load(mock_by_ids.calls[0].request.body)

    def test_standard_retrieve_multiple_fail(self, api_client_with_api_key, rsps):
        rsps.add(rsps.POST, BASE_URL + URL_PATH + "/byids", status=400, json={"error": {"message": "Client Error"}})
        with pytest.raises(CogniteAPIError, match="Client Error") as e:
            api_client_with_api_key._retrieve_multiple(
                cls=SomeResourceList, resource_path=URL_PATH, wrap_ids=True, ids=[1, 2]
            )
        assert "Client Error" == e.value.message
        assert 400 == e.value.code

    def test_ids_all_None(self, api_client_with_api_key):
        with pytest.raises(ValueError, match="No ids specified"):
            api_client_with_api_key._retrieve_multiple(cls=SomeResourceList, resource_path=URL_PATH, wrap_ids=False)

    def test_single_id_not_found(self, api_client_with_api_key, rsps):
        rsps.add(
            rsps.POST,
            BASE_URL + URL_PATH + "/byids",
            status=400,
            json={"error": {"message": "Not Found", "missing": [{"id": 1}]}},
        )
        res = api_client_with_api_key._retrieve_multiple(
            cls=SomeResourceList, resource_path=URL_PATH, wrap_ids=True, ids=1
        )
        assert res is None

    def test_multiple_ids_not_found(self, api_client_with_api_key, rsps):
        rsps.add(
            rsps.POST,
            BASE_URL + URL_PATH + "/byids",
            status=400,
            json={"error": {"message": "Not Found", "missing": [{"id": 1}]}},
        )
        rsps.add(
            rsps.POST,
            BASE_URL + URL_PATH + "/byids",
            status=400,
            json={"error": {"message": "Not Found", "missing": [{"id": 2}]}},
        )
        with set_request_limit(api_client_with_api_key, 1):
            with pytest.raises(CogniteNotFoundError) as e:
                api_client_with_api_key._retrieve_multiple(
                    cls=SomeResourceList, resource_path=URL_PATH, wrap_ids=True, ids=[1, 2]
                )
        assert [{"id": 1}, {"id": 2}] == e.value.not_found

    def test_cognite_client_is_set(self, cognite_client, api_client_with_api_key, mock_by_ids):
        assert (
            cognite_client
            == api_client_with_api_key._retrieve_multiple(
                cls=SomeResourceList, resource_path=URL_PATH, wrap_ids=True, ids=[1, 2]
            )._cognite_client
        )

    def test_over_limit_concurrent(self, api_client_with_api_key, rsps):
        rsps.add(rsps.POST, BASE_URL + URL_PATH + "/byids", status=200, json={"items": [{"x": 1, "y": 2}]})
        rsps.add(rsps.POST, BASE_URL + URL_PATH + "/byids", status=200, json={"items": [{"x": 3, "y": 4}]})

        with set_request_limit(api_client_with_api_key, 1):
            api_client_with_api_key._retrieve_multiple(
                cls=SomeResourceList, resource_path=URL_PATH, ids=[1, 2], wrap_ids=False
            )

        assert {"items": [1]} == jsgz_load(rsps.calls[0].request.body)
        assert {"items": [2]} == jsgz_load(rsps.calls[1].request.body)


class TestStandardList:
    def test_standard_list_ok(self, api_client_with_api_key, rsps):
        rsps.add(rsps.GET, BASE_URL + URL_PATH, status=200, json={"items": [{"x": 1, "y": 2}, {"x": 1}]})
        assert (
            SomeResourceList([SomeResource(1, 2), SomeResource(1)]).dump()
            == api_client_with_api_key._list(cls=SomeResourceList, resource_path=URL_PATH, method="GET").dump()
        )

    def test_standard_list_with_filter_GET_ok(self, api_client_with_api_key, rsps):
        rsps.add(rsps.GET, BASE_URL + URL_PATH, status=200, json={"items": [{"x": 1, "y": 2}, {"x": 1}]})
        assert (
            SomeResourceList([SomeResource(1, 2), SomeResource(1)]).dump()
            == api_client_with_api_key._list(
                cls=SomeResourceList, resource_path=URL_PATH, method="GET", filter={"filter": "bla"}
            ).dump()
        )
        assert "filter=bla" in rsps.calls[0].request.path_url

    def test_standard_list_with_filter_POST_ok(self, api_client_with_api_key, rsps):
        rsps.add(rsps.POST, BASE_URL + URL_PATH + "/list", status=200, json={"items": [{"x": 1, "y": 2}, {"x": 1}]})
        assert SomeResourceList([SomeResource(1, 2), SomeResource(1)]) == api_client_with_api_key._list(
            cls=SomeResourceList, resource_path=URL_PATH, method="POST", filter={"filter": "bla"}
        )
        assert {"filter": {"filter": "bla"}, "limit": 1000, "cursor": None} == jsgz_load(rsps.calls[0].request.body)

    def test_standard_list_fail(self, api_client_with_api_key, rsps):
        rsps.add(rsps.GET, BASE_URL + URL_PATH, status=400, json={"error": {"message": "Client Error"}})
        with pytest.raises(CogniteAPIError, match="Client Error") as e:
            api_client_with_api_key._list(cls=SomeResourceList, resource_path=URL_PATH, method="GET")
        assert 400 == e.value.code
        assert "Client Error" == e.value.message

    NUMBER_OF_ITEMS_FOR_AUTOPAGING = 11500
    ITEMS_TO_GET_WHILE_AUTOPAGING = [{"x": 1, "y": 1} for _ in range(NUMBER_OF_ITEMS_FOR_AUTOPAGING)]

    def test_list_partitions(self, api_client_with_api_key, rsps):
        rsps.add(rsps.POST, BASE_URL + URL_PATH + "/list", status=200, json={"items": [{"x": 1, "y": 2}, {"x": 1}]})
        res = api_client_with_api_key._list(
            cls=SomeResourceList,
            resource_path=URL_PATH,
            method="POST",
            partitions=3,
            limit=None,
            headers={"X-Test": "foo"},
        )
        assert 6 == len(res)
        assert isinstance(res, SomeResourceList)
        assert isinstance(res[0], SomeResource)
        assert 3 == len(rsps.calls)
        assert {"1/3", "2/3", "3/3"} == {jsgz_load(c.request.body)["partition"] for c in rsps.calls}
        for call in rsps.calls:
            request = jsgz_load(call.request.body)
            assert "X-Test" in call.request.headers.keys()
            del request["partition"]
            assert {"cursor": None, "filter": {}, "limit": 1000} == request
            assert call.response.json()["items"] == [{"x": 1, "y": 2}, {"x": 1}]

    def test_list_partitions_with_failure(self, api_client_with_api_key, rsps):
        def request_callback(request):
            payload = jsgz_load(request.body)
            np, total = payload["partition"].split("/")
            if int(np) == 2:
                return 503, {}, json.dumps({"message": "Service Unavailable"})
            else:
                return 200, {}, json.dumps({"items": [{"x": 42, "y": 13}]})

        rsps.add_callback(
            rsps.POST, BASE_URL + URL_PATH + "/list", callback=request_callback, content_type="application/json"
        )
        with pytest.raises(CogniteAPIError) as exc:
            api_client_with_api_key._list(
                cls=SomeResourceList, resource_path=URL_PATH, method="POST", partitions=4, limit=None
            )
        assert 503 == exc.value.code
        assert 4 == len(rsps.calls)

    @pytest.fixture
    def mock_get_for_autopaging(self, rsps):
        def callback(request):
            params = {elem.split("=")[0]: elem.split("=")[1] for elem in request.path_url.split("?")[-1].split("&")}
            limit = int(params["limit"])
            cursor = int(params.get("cursor") or 0)
            items = self.ITEMS_TO_GET_WHILE_AUTOPAGING[cursor : cursor + limit]
            if cursor + limit >= self.NUMBER_OF_ITEMS_FOR_AUTOPAGING:
                next_cursor = None
            else:
                next_cursor = cursor + limit
            response = json.dumps({"nextCursor": next_cursor, "items": items})
            return 200, {}, response

        rsps.add_callback(rsps.GET, BASE_URL + URL_PATH, callback)

    @pytest.fixture
    def mock_get_for_autopaging_2589(self, rsps):
        NUM_ITEMS = 2589
        ITEMS_EDGECASE = [{"x": 1, "y": 1} for _ in range(NUM_ITEMS)]

        def callback(request):
            params = {elem.split("=")[0]: elem.split("=")[1] for elem in request.path_url.split("?")[-1].split("&")}
            limit = int(params["limit"])
            cursor = int(params.get("cursor") or 0)
            items = ITEMS_EDGECASE[cursor : cursor + limit]
            if cursor + limit >= NUM_ITEMS:
                next_cursor = None
            else:
                next_cursor = cursor + limit
            response = json.dumps({"nextCursor": next_cursor, "items": items})
            return 200, {}, response

        rsps.add_callback(rsps.GET, BASE_URL + URL_PATH, callback)

    @pytest.mark.usefixtures("mock_get_for_autopaging")
    def test_standard_list_generator(self, api_client_with_api_key):
        total_resources = 0
        for resource in api_client_with_api_key._list_generator(
            cls=SomeResourceList, resource_path=URL_PATH, method="GET"
        ):
            assert isinstance(resource, SomeResource)
            total_resources += 1
        assert 11500 == total_resources

    @pytest.mark.usefixtures("mock_get_for_autopaging")
    def test_standard_list_generator_with_limit(self, api_client_with_api_key):
        total_resources = 0
        for resource in api_client_with_api_key._list_generator(
            cls=SomeResourceList, resource_path=URL_PATH, method="GET", limit=10000
        ):
            assert isinstance(resource, SomeResource)
            total_resources += 1
        assert 10000 == total_resources

    @pytest.mark.usefixtures("mock_get_for_autopaging")
    def test_standard_list_generator_with_chunk_size(self, api_client_with_api_key):
        total_resources = 0
        for resource_chunk in api_client_with_api_key._list_generator(
            cls=SomeResourceList, resource_path=URL_PATH, method="GET", chunk_size=1000
        ):
            assert isinstance(resource_chunk, SomeResourceList)
            if len(resource_chunk) == 1000:
                total_resources += 1000
            elif len(resource_chunk) == 500:
                total_resources += 500
            else:
                raise AssertionError("resource chunk length was not 1000 or 500")
        assert 11500 == total_resources

    @pytest.mark.usefixtures("mock_get_for_autopaging_2589")
    def test_standard_list_generator_with_chunk_size_chunk_edge_case(self, api_client_with_api_key):
        total_resources = 0
        for resource_chunk in api_client_with_api_key._list_generator(
            cls=SomeResourceList, resource_path=URL_PATH, method="GET", chunk_size=2500
        ):
            assert isinstance(resource_chunk, SomeResourceList)
            total_resources += len(resource_chunk)
            assert len(resource_chunk) in [89, 2500]
        assert 2589 == total_resources

    @pytest.mark.usefixtures("mock_get_for_autopaging_2589")
    def test_standard_list_generator_with_chunk_size_chunk_limit(self, api_client_with_api_key):
        total_resources = 0
        for resource_chunk in api_client_with_api_key._list_generator(
            cls=SomeResourceList, resource_path=URL_PATH, method="GET", chunk_size=2500, limit=2563
        ):
            assert isinstance(resource_chunk, SomeResourceList)
            total_resources += len(resource_chunk)
            assert len(resource_chunk) in [63, 2500]
        assert 2563 == total_resources

    @pytest.mark.usefixtures("mock_get_for_autopaging")
    def test_standard_list_generator_with_chunk_size_with_limit(self, api_client_with_api_key):
        total_resources = 0
        for resource_chunk in api_client_with_api_key._list_generator(
            cls=SomeResourceList, resource_path=URL_PATH, method="GET", limit=10000, chunk_size=1000
        ):
            assert isinstance(resource_chunk, SomeResourceList)
            assert 1000 == len(resource_chunk)
            total_resources += 1000
        assert 10000 == total_resources

    @pytest.mark.usefixtures("mock_get_for_autopaging")
    def test_standard_list_generator__chunk_size_exceeds_max(self, api_client_with_api_key):
        total_resources = 0
        for resource_chunk in api_client_with_api_key._list_generator(
            cls=SomeResourceList, resource_path=URL_PATH, method="GET", limit=2002, chunk_size=1001
        ):
            assert isinstance(resource_chunk, SomeResourceList)
            assert 1001 == len(resource_chunk)
            total_resources += 1001
        assert 2002 == total_resources

    @pytest.mark.usefixtures("mock_get_for_autopaging")
    def test_standard_list_autopaging(self, api_client_with_api_key):
        res = api_client_with_api_key._list(cls=SomeResourceList, resource_path=URL_PATH, method="GET")
        assert self.NUMBER_OF_ITEMS_FOR_AUTOPAGING == len(res)

    @pytest.mark.usefixtures("mock_get_for_autopaging")
    def test_standard_list_autopaging_with_limit(self, api_client_with_api_key):
        res = api_client_with_api_key._list(cls=SomeResourceList, resource_path=URL_PATH, method="GET", limit=5333)
        assert 5333 == len(res)

    def test_cognite_client_is_set(self, cognite_client, api_client_with_api_key, rsps):
        rsps.add(rsps.POST, BASE_URL + URL_PATH + "/list", status=200, json={"items": [{"x": 1, "y": 2}, {"x": 1}]})
        rsps.add(rsps.GET, BASE_URL + URL_PATH, status=200, json={"items": [{"x": 1, "y": 2}, {"x": 1}]})
        assert (
            cognite_client
            == api_client_with_api_key._list(
                cls=SomeResourceList, resource_path=URL_PATH, method="POST"
            )._cognite_client
        )
        assert (
            cognite_client
            == api_client_with_api_key._list(cls=SomeResourceList, resource_path=URL_PATH, method="GET")._cognite_client
        )


class TestStandardAggregate:
    def test_standard_aggregate_OK(self, api_client_with_api_key, rsps):
        rsps.add(rsps.POST, BASE_URL + URL_PATH + "/aggregate", status=200, json={"items": [{"count": 1}]})
        assert [SomeAggregation(1)] == api_client_with_api_key._aggregate(
            resource_path=URL_PATH, filter={"x": 1}, cls=SomeAggregation
        )

    def test_standard_aggregate_fail(self, api_client_with_api_key, rsps):
        rsps.add(rsps.POST, BASE_URL + URL_PATH + "/aggregate", status=400, json={"error": {"message": "Client Error"}})
        with pytest.raises(CogniteAPIError, match="Client Error") as e:
            api_client_with_api_key._aggregate(resource_path=URL_PATH, filter={"x": 1}, cls=SomeAggregation)
        assert "Client Error" == e.value.message
        assert 400 == e.value.code


class TestStandardCreate:
    def test_standard_create_ok(self, api_client_with_api_key, rsps):
        rsps.add(rsps.POST, BASE_URL + URL_PATH, status=200, json={"items": [{"x": 1, "y": 2}, {"x": 1}]})
        res = api_client_with_api_key._create_multiple(
            cls=SomeResourceList, resource_path=URL_PATH, items=[SomeResource(1, 1), SomeResource(1)]
        )
        assert {"items": [{"x": 1, "y": 1}, {"x": 1}]} == jsgz_load(rsps.calls[0].request.body)
        assert SomeResource(1, 2) == res[0]
        assert SomeResource(1) == res[1]

    def test_standard_create_extra_body_fields(self, api_client_with_api_key, rsps):
        rsps.add(rsps.POST, BASE_URL + URL_PATH, status=200, json={"items": [{"x": 1, "y": 2}, {"x": 1}]})
        res = api_client_with_api_key._create_multiple(
            cls=SomeResourceList,
            resource_path=URL_PATH,
            items=[SomeResource(1, 1), SomeResource(1)],
            extra_body_fields={"foo": "bar"},
        )
        assert {"items": [{"x": 1, "y": 1}, {"x": 1}], "foo": "bar"} == jsgz_load(rsps.calls[0].request.body)

    def test_standard_create_single_item_ok(self, api_client_with_api_key, rsps):
        rsps.add(rsps.POST, BASE_URL + URL_PATH, status=200, json={"items": [{"x": 1, "y": 2}]})
        res = api_client_with_api_key._create_multiple(
            cls=SomeResourceList, resource_path=URL_PATH, items=SomeResource(1, 2)
        )
        assert {"items": [{"x": 1, "y": 2}]} == jsgz_load(rsps.calls[0].request.body)
        assert SomeResource(1, 2) == res

    def test_standard_create_single_item_in_list_ok(self, api_client_with_api_key, rsps):
        rsps.add(rsps.POST, BASE_URL + URL_PATH, status=200, json={"items": [{"x": 1, "y": 2}]})
        res = api_client_with_api_key._create_multiple(
            cls=SomeResourceList, resource_path=URL_PATH, items=[SomeResource(1, 2)]
        )
        assert {"items": [{"x": 1, "y": 2}]} == jsgz_load(rsps.calls[0].request.body)
        assert SomeResourceList([SomeResource(1, 2)]) == res

    def test_standard_create_fail(self, api_client_with_api_key, rsps):
        def callback(request):
            item = jsgz_load(request.body)["items"][0]
            return int(item["externalId"]), {}, json.dumps({})

        rsps.add_callback(rsps.POST, BASE_URL + URL_PATH, callback=callback, content_type="application/json")
        with set_request_limit(api_client_with_api_key, 1):
            with pytest.raises(CogniteAPIError) as e:
                api_client_with_api_key._create_multiple(
                    cls=SomeResourceList,
                    resource_path=URL_PATH,
                    items=[
                        SomeResource(1, external_id="400"),
                        SomeResource(external_id="500"),
                        SomeResource(1, 1, external_id="200"),
                    ],
                )
        assert 500 == e.value.code
        assert [SomeResource(1, external_id="400")] == e.value.failed
        assert [SomeResource(1, 1, external_id="200")] == e.value.successful
        assert [SomeResource(external_id="500")] == e.value.unknown

    def test_standard_create_concurrent(self, api_client_with_api_key, rsps):
        rsps.add(rsps.POST, BASE_URL + URL_PATH, status=200, json={"items": [{"x": 1, "y": 2}]})
        rsps.add(rsps.POST, BASE_URL + URL_PATH, status=200, json={"items": [{"x": 3, "y": 4}]})

        res = api_client_with_api_key._create_multiple(
            cls=SomeResourceList, resource_path=URL_PATH, items=[SomeResource(1, 2), SomeResource(3, 4)], limit=1
        )
        assert SomeResourceList([SomeResource(1, 2), SomeResource(3, 4)]) == res

        assert {"items": [{"x": 1, "y": 2}]} == jsgz_load(rsps.calls[0].request.body)
        assert {"items": [{"x": 3, "y": 4}]} == jsgz_load(rsps.calls[1].request.body)

    def test_cognite_client_is_set(self, cognite_client, api_client_with_api_key, rsps):
        rsps.add(rsps.POST, BASE_URL + URL_PATH, status=200, json={"items": [{"x": 1, "y": 2}]})
        assert (
            cognite_client
            == api_client_with_api_key._create_multiple(
                cls=SomeResourceList, resource_path=URL_PATH, items=SomeResource()
            )._cognite_client
        )
        assert (
            cognite_client
            == api_client_with_api_key._create_multiple(
                cls=SomeResourceList, resource_path=URL_PATH, items=[SomeResource()]
            )._cognite_client
        )


class TestStandardDelete:
    def test_standard_delete_multiple_ok(self, api_client_with_api_key, rsps):
        rsps.add(rsps.POST, BASE_URL + URL_PATH + "/delete", status=200, json={})
        api_client_with_api_key._delete_multiple(resource_path=URL_PATH, wrap_ids=False, ids=[1, 2])
        assert {"items": [1, 2]} == jsgz_load(rsps.calls[0].request.body)

    def test_standard_delete_multiple_ok__single_id(self, api_client_with_api_key, rsps):
        rsps.add(rsps.POST, BASE_URL + URL_PATH + "/delete", status=200, json={})
        api_client_with_api_key._delete_multiple(resource_path=URL_PATH, wrap_ids=False, ids=1)
        assert {"items": [1]} == jsgz_load(rsps.calls[0].request.body)

    def test_standard_delete_multiple_ok__single_id_in_list(self, api_client_with_api_key, rsps):
        rsps.add(rsps.POST, BASE_URL + URL_PATH + "/delete", status=200, json={})
        api_client_with_api_key._delete_multiple(resource_path=URL_PATH, wrap_ids=False, ids=[1])
        assert {"items": [1]} == jsgz_load(rsps.calls[0].request.body)

    def test_standard_delete_multiple_fail_4xx(self, api_client_with_api_key, rsps):
        rsps.add(rsps.POST, BASE_URL + URL_PATH + "/delete", status=400, json={"error": {"message": "Client Error"}})
        with pytest.raises(CogniteAPIError) as e:
            api_client_with_api_key._delete_multiple(resource_path=URL_PATH, wrap_ids=False, ids=[1, 2])
        assert 400 == e.value.code
        assert "Client Error" == e.value.message
        assert e.value.failed == [1, 2]

    def test_standard_delete_multiple_fail_5xx(self, api_client_with_api_key, rsps):
        rsps.add(rsps.POST, BASE_URL + URL_PATH + "/delete", status=500, json={"error": {"message": "Server Error"}})
        with pytest.raises(CogniteAPIError) as e:
            api_client_with_api_key._delete_multiple(resource_path=URL_PATH, wrap_ids=False, ids=[1, 2])
        assert 500 == e.value.code
        assert "Server Error" == e.value.message
        assert e.value.unknown == [1, 2]
        assert e.value.failed == []

    def test_standard_delete_multiple_fail_missing_ids(self, api_client_with_api_key, rsps):
        rsps.add(
            rsps.POST,
            BASE_URL + URL_PATH + "/delete",
            status=400,
            json={"error": {"message": "Missing ids", "missing": [{"id": 1}]}},
        )
        rsps.add(
            rsps.POST,
            BASE_URL + URL_PATH + "/delete",
            status=400,
            json={"error": {"message": "Missing ids", "missing": [{"id": 3}]}},
        )
        with set_request_limit(api_client_with_api_key, 2):
            with pytest.raises(CogniteNotFoundError) as e:
                api_client_with_api_key._delete_multiple(resource_path=URL_PATH, wrap_ids=False, ids=[1, 2, 3])

        assert [{"id": 1}, {"id": 3}] == e.value.not_found
        assert [1, 2, 3] == e.value.failed

    def test_over_limit_concurrent(self, api_client_with_api_key, rsps):
        rsps.add(rsps.POST, BASE_URL + URL_PATH + "/delete", status=200, json={})
        rsps.add(rsps.POST, BASE_URL + URL_PATH + "/delete", status=200, json={})

        with set_request_limit(api_client_with_api_key, 2):
            api_client_with_api_key._delete_multiple(resource_path=URL_PATH, ids=[1, 2, 3, 4], wrap_ids=False)
        assert {"items": [1, 2]} == jsgz_load(rsps.calls[0].request.body)
        assert {"items": [3, 4]} == jsgz_load(rsps.calls[1].request.body)


class TestStandardUpdate:
    @pytest.fixture
    def mock_update(self, api_client_with_api_key, rsps):
        rsps.add(rsps.POST, BASE_URL + URL_PATH + "/update", status=200, json={"items": [{"id": 1, "x": 1, "y": 100}]})
        yield rsps

    def test_standard_update_with_cognite_resource_OK(self, api_client_with_api_key, mock_update):
        res = api_client_with_api_key._update_multiple(
            cls=SomeResourceList, resource_path=URL_PATH, items=[SomeResource(id=1, y=100)]
        )
        assert SomeResourceList([SomeResource(id=1, x=1, y=100)]) == res
        assert {"items": [{"id": 1, "update": {"y": {"set": 100}}}]} == jsgz_load(mock_update.calls[0].request.body)

    def test_standard_update_with_cognite_resource__subject_to_camel_case_issue(
        self, api_client_with_api_key, mock_update
    ):
        api_client_with_api_key._update_multiple(
            cls=SomeResourceList, resource_path=URL_PATH, items=[SomeResource(id=1, external_id="abc", y=100)]
        )
        assert {"items": [{"id": 1, "update": {"y": {"set": 100}, "externalId": {"set": "abc"}}}]} == jsgz_load(
            mock_update.calls[0].request.body
        )

    def test_standard_update_with_cognite_resource__non_update_attributes(self, api_client_with_api_key, mock_update):
        res = api_client_with_api_key._update_multiple(
            cls=SomeResourceList, resource_path=URL_PATH, items=[SomeResource(id=1, y=100, x=1)]
        )
        assert SomeResourceList([SomeResource(id=1, x=1, y=100)]) == res
        assert {"items": [{"id": 1, "update": {"y": {"set": 100}}}]} == jsgz_load(mock_update.calls[0].request.body)

    def test_standard_update_with_cognite_resource__id_and_external_id_set(self, api_client_with_api_key, mock_update):
        api_client_with_api_key._update_multiple(
            cls=SomeResourceList, resource_path=URL_PATH, items=[SomeResource(id=1, external_id="1", y=100, x=1)]
        )
        assert {"items": [{"id": 1, "update": {"y": {"set": 100}, "externalId": {"set": "1"}}}]} == jsgz_load(
            mock_update.calls[0].request.body
        )

    def test_standard_update_with_cognite_resource_and_external_id_OK(self, api_client_with_api_key, mock_update):
        res = api_client_with_api_key._update_multiple(
            cls=SomeResourceList, resource_path=URL_PATH, items=[SomeResource(external_id="1", y=100)]
        )
        assert SomeResourceList([SomeResource(id=1, x=1, y=100)]) == res
        assert {"items": [{"externalId": "1", "update": {"y": {"set": 100}}}]} == jsgz_load(
            mock_update.calls[0].request.body
        )

    def test_standard_update_with_cognite_update_object_OK(self, api_client_with_api_key, mock_update):
        res = api_client_with_api_key._update_multiple(
            cls=SomeResourceList, resource_path=URL_PATH, items=[SomeUpdate(id=1).y.set(100)]
        )
        assert SomeResourceList([SomeResource(id=1, x=1, y=100)]) == res
        assert {"items": [{"id": 1, "update": {"y": {"set": 100}}}]} == jsgz_load(mock_update.calls[0].request.body)

    def test_standard_update_single_object(self, api_client_with_api_key, mock_update):
        res = api_client_with_api_key._update_multiple(
            cls=SomeResourceList, resource_path=URL_PATH, items=SomeUpdate(id=1).y.set(100)
        )
        assert SomeResource(id=1, x=1, y=100) == res
        assert {"items": [{"id": 1, "update": {"y": {"set": 100}}}]} == jsgz_load(mock_update.calls[0].request.body)

    def test_standard_update_with_cognite_update_object_and_external_id_OK(self, api_client_with_api_key, mock_update):
        res = api_client_with_api_key._update_multiple(
            cls=SomeResourceList, resource_path=URL_PATH, items=[SomeUpdate(external_id="1").y.set(100)]
        )
        assert SomeResourceList([SomeResource(id=1, x=1, y=100)]) == res
        assert {"items": [{"externalId": "1", "update": {"y": {"set": 100}}}]} == jsgz_load(
            mock_update.calls[0].request.body
        )

    def test_standard_update_fail_4xx(self, api_client_with_api_key, rsps):
        rsps.add(rsps.POST, BASE_URL + URL_PATH + "/update", status=400, json={"error": {"message": "Client Error"}})
        with pytest.raises(CogniteAPIError) as e:
            api_client_with_api_key._update_multiple(
                cls=SomeResourceList,
                resource_path=URL_PATH,
                items=[SomeResource(id=0), SomeResource(external_id="abc")],
            )
        assert e.value.message == "Client Error"
        assert e.value.code == 400
        assert e.value.failed == [0, "abc"]

    def test_standard_update_fail_5xx(self, api_client_with_api_key, rsps):
        rsps.add(rsps.POST, BASE_URL + URL_PATH + "/update", status=500, json={"error": {"message": "Server Error"}})
        with pytest.raises(CogniteAPIError) as e:
            api_client_with_api_key._update_multiple(
                cls=SomeResourceList,
                resource_path=URL_PATH,
                items=[SomeResource(id=0), SomeResource(external_id="abc")],
            )
        assert e.value.message == "Server Error"
        assert e.value.code == 500
        assert e.value.failed == []
        assert e.value.unknown == [0, "abc"]

    def test_standard_update_fail_missing_and_5xx(self, api_client_with_api_key, rsps):
        rsps.add(
            rsps.POST,
            BASE_URL + URL_PATH + "/update",
            status=400,
            json={"error": {"message": "Missing ids", "missing": [{"id": 0}]}},
        )
        rsps.add(rsps.POST, BASE_URL + URL_PATH + "/update", status=500, json={"error": {"message": "Server Error"}})
        with set_request_limit(api_client_with_api_key, 1):
            with pytest.raises(CogniteAPIError) as e:
                api_client_with_api_key._update_multiple(
                    cls=SomeResourceList,
                    resource_path=URL_PATH,
                    items=[SomeResource(id=0), SomeResource(external_id="abc")],
                )
        assert ["abc"] == e.value.unknown
        assert [0] == e.value.failed
        assert [{"id": 0}] == e.value.missing

    def test_cognite_client_is_set(self, cognite_client, api_client_with_api_key, mock_update):
        assert (
            cognite_client
            == api_client_with_api_key._update_multiple(
                cls=SomeResourceList, resource_path=URL_PATH, items=SomeResource(id=0)
            )._cognite_client
        )
        assert (
            cognite_client
            == api_client_with_api_key._update_multiple(
                cls=SomeResourceList, resource_path=URL_PATH, items=[SomeResource(id=0)]
            )._cognite_client
        )

    def test_over_limit_concurrent(self, api_client_with_api_key, rsps):
        rsps.add(rsps.POST, BASE_URL + URL_PATH + "/update", status=200, json={"items": [{"x": 1, "y": 2}]})
        rsps.add(rsps.POST, BASE_URL + URL_PATH + "/update", status=200, json={"items": [{"x": 3, "y": 4}]})

        with set_request_limit(api_client_with_api_key, 1):
            api_client_with_api_key._update_multiple(
                cls=SomeResourceList, resource_path=URL_PATH, items=[SomeResource(1, 2, id=1), SomeResource(3, 4, id=2)]
            )

        assert {"items": [{"id": 1, "update": {"y": {"set": 2}}}]} == jsgz_load(rsps.calls[0].request.body)
        assert {"items": [{"id": 2, "update": {"y": {"set": 4}}}]} == jsgz_load(rsps.calls[1].request.body)


class TestStandardSearch:
    def test_standard_search_ok(self, api_client_with_api_key, rsps):
        rsps.add(rsps.POST, BASE_URL + URL_PATH + "/search", status=200, json={"items": [{"x": 1, "y": 2}]})

        res = api_client_with_api_key._search(
            cls=SomeResourceList,
            resource_path=URL_PATH,
            search={"name": "bla"},
            filter=SomeFilter(var_x=1, var_y=1),
            limit=1000,
        )
        assert SomeResourceList([SomeResource(1, 2)]) == res
        assert {"search": {"name": "bla"}, "limit": 1000, "filter": {"varX": 1, "varY": 1}} == jsgz_load(
            rsps.calls[0].request.body
        )

    def test_standard_search_dict_filter_ok(self, api_client_with_api_key, rsps):
        rsps.add(rsps.POST, BASE_URL + URL_PATH + "/search", status=200, json={"items": [{"x": 1, "y": 2}]})

        res = api_client_with_api_key._search(
            cls=SomeResourceList,
            resource_path=URL_PATH,
            search={"name": "bla"},
            filter={"var_x": 1, "varY": 1},
            limit=1000,
        )
        assert SomeResourceList([SomeResource(1, 2)]) == res
        assert {"search": {"name": "bla"}, "limit": 1000, "filter": {"varX": 1, "varY": 1}} == jsgz_load(
            rsps.calls[0].request.body
        )

    def test_standard_search_fail(self, api_client_with_api_key, rsps):
        rsps.add(rsps.POST, BASE_URL + URL_PATH + "/search", status=400, json={"error": {"message": "Client Error"}})

        with pytest.raises(CogniteAPIError, match="Client Error") as e:
            api_client_with_api_key._search(
                cls=SomeResourceList, resource_path=URL_PATH, search=None, filter=None, limit=None
            )
        assert "Client Error" == e.value.message
        assert 400 == e.value.code

    def test_cognite_client_is_set(self, cognite_client, api_client_with_api_key, rsps):
        rsps.add(rsps.POST, BASE_URL + URL_PATH + "/search", status=200, json={"items": [{"x": 1, "y": 2}]})

        assert (
            cognite_client
            == api_client_with_api_key._search(
                cls=SomeResourceList, resource_path=URL_PATH, search={"name": "bla"}, filter={"name": "bla"}, limit=1000
            )._cognite_client
        )


class TestHelpers:
    @pytest.mark.parametrize(
        "ids, external_ids, wrap_ids, expected",
        [
            (1, None, False, [1]),
            ([1, 2], None, False, [1, 2]),
            (1, None, True, [{"id": 1}]),
            ([1, 2], None, True, [{"id": 1}, {"id": 2}]),
            (1, "1", True, [{"id": 1}, {"externalId": "1"}]),
            (1, ["1"], True, [{"id": 1}, {"externalId": "1"}]),
            ([1, 2], ["1"], True, [{"id": 1}, {"id": 2}, {"externalId": "1"}]),
            (None, "1", True, [{"externalId": "1"}]),
            (None, ["1", "2"], True, [{"externalId": "1"}, {"externalId": "2"}]),
        ],
    )
    def test_process_ids(self, api_client_with_api_key, ids, external_ids, wrap_ids, expected):
        assert expected == api_client_with_api_key._process_ids(ids, external_ids, wrap_ids)

    @pytest.mark.parametrize(
        "ids, external_ids, wrap_ids, exception, match",
        [
            (None, None, False, ValueError, "No ids specified"),
            (None, ["1", "2"], False, ValueError, "externalIds must be wrapped"),
            ([1], ["1"], False, ValueError, "externalIds must be wrapped"),
            ("1", None, False, TypeError, "must be int or list of int"),
            (1, 1, True, TypeError, "must be str or list of str"),
        ],
    )
    def test_process_ids_fail(self, api_client_with_api_key, ids, external_ids, wrap_ids, exception, match):
        with pytest.raises(exception, match=match):
            api_client_with_api_key._process_ids(ids, external_ids, wrap_ids)

    @pytest.mark.parametrize(
        "id, external_id, expected",
        [(1, None, True), (None, "1", True), (None, None, False), ([1], None, False), (None, ["1"], False)],
    )
    def test_is_single_identifier(self, api_client_with_api_key, id, external_id, expected):
        assert expected == api_client_with_api_key._is_single_identifier(id, external_id)

    @pytest.mark.parametrize(
        "method, path, expected",
        [
            ("GET", "https://api.cognitedata.com/login/status", True),
            ("GET", "https://greenfield.cognitedata.com/api/v1/projects/blabla/assets", True),
            ("POST", "https://localhost:8000/api/v1/projects/blabla/files/list", True),
            ("PUT", "https://api.cognitedata.com/bla", True),
            (
                "POST",
                "https://api.cognitedata.com/api/v1/projects/sebnickelgreenfield/files/downloadlink?extendedExpiration=true",
                True,
            ),
            ("POST", "https://api.cognitedata.com/api/v1/projects/blabla/timeseries/list", True),
            ("POST", "https://greenfield.cognitedata.com/api/v1/projects/blabla/assets", False),
            ("POST", "https://greenfield.cognitedata.com/api/playground/projects/blabla/relationships/list", True),
            ("PUT", "https://localhost:8000.com/api/v1/projects/blabla/assets", True),
            ("PATCH", "https://localhost:8000.com/api/v1/projects/blabla/patchy", True),
            ("GET", "https://another-cluster.cognitedata.com/login/status", True),
        ],
    )
    def test_is_retryable(self, api_client_with_api_key, method, path, expected):
        assert expected == api_client_with_api_key._is_retryable(method, path)

    @pytest.mark.parametrize(
        "method, path", [("POST", "htt://bla/bla"), ("BLOP", "http://localhost:8000/login/status")]
    )
    def test_is_retryable_fail(self, api_client_with_api_key, method, path):
        with pytest.raises(ValueError, match="is not valid"):
            api_client_with_api_key._is_retryable(method, path)

    def test_is_retryable_add(self, api_client_with_api_key):
        assert False == api_client_with_api_key._is_retryable(
            "POST", "https://greenfield.cognitedata.com/api/v1/projects/blabla/assets/bloop"
        )
        APIClient.RETRYABLE_POST_ENDPOINTS.add("/assets/bloop")
        assert True == api_client_with_api_key._is_retryable(
            "POST", "https://greenfield.cognitedata.com/api/v1/projects/blabla/assets/bloop"
        )
        APIClient.RETRYABLE_POST_ENDPOINTS.remove("/assets/bloop")
        assert False == api_client_with_api_key._is_retryable(
            "POST", "https://greenfield.cognitedata.com/api/v1/projects/blabla/assets/bloop"
        )

    def test_get_status_codes_to_retry(self):
        os.environ["COGNITE_STATUS_FORCELIST"] = "1,2, 3,4"
        assert [1, 2, 3, 4] == utils._client_config._DefaultConfig().status_forcelist
        del os.environ["COGNITE_STATUS_FORCELIST"]

    @pytest.mark.parametrize(
        "before, after",
        [
            ({"api-key": "bla", "key": "bla"}, {"api-key": "***", "key": "bla"}),
            ({"Authorization": "bla", "key": "bla"}, {"Authorization": "***", "key": "bla"}),
        ],
    )
    def test_sanitize_headers(self, before, after):
        assert before != after
        APIClient._sanitize_headers(before)
        assert before == after

    @pytest.mark.parametrize(
        "content, expected",
        [('{"foo": 42}'.encode(), '{"foo": 42}'), ("foobar".encode(), "foobar"), (b"\xed\xbc\xad", "<binary>")],
    )
    def test_get_response_content_safe(self, content, expected):
        res = Response()
        res._content = content
        assert APIClient._get_response_content_safe(res) == expected


class TestConnectionPooling:
    def test_connection_pool_is_shared_between_clients(self):
        with set_env_var("COGNITE_API_KEY", "bla"):
            c1 = CogniteClient()
            c2 = CogniteClient()
        assert (
            c1._api_client._http_client.session
            == c2._api_client._http_client.session
            == c1._api_client._http_client_with_retry.session
            == c2._api_client._http_client_with_retry.session
        )



import random
import string

import pytest

from cognite.client.data_classes import Transformation, TransformationDestination, TransformationNotification

prefix = "".join(random.choice(string.ascii_letters) for i in range(6))


@pytest.fixture
def new_transformation(cognite_client):
    transform = Transformation(
        external_id=f"newTransformation-{prefix}",
        name="any",
        destination=TransformationDestination.assets(),
        query="select * from _cdf.assets",
        source_api_key=cognite_client.config.api_key,
        destination_api_key=cognite_client.config.api_key,
    )
    ts = cognite_client.transformations.create(transform)

    yield ts

    cognite_client.transformations.delete(id=ts.id)
    assert cognite_client.transformations.retrieve(ts.id) is None


@pytest.fixture
def new_notification(cognite_client, new_transformation):
    notification = TransformationNotification(
        transformation_id=new_transformation.id, destination=f"my_{prefix}@email.com"
    )
    tn = cognite_client.transformations.notifications.create(notification)

    yield tn

    cognite_client.transformations.notifications.delete(id=tn.id)
    assert len(cognite_client.transformations.notifications.list(transformation_id=new_transformation.id)) == 0
    assert tn.transformation_id == new_transformation.id


@pytest.fixture
def new_notification_by_external_id(cognite_client, new_transformation):
    notification = TransformationNotification(
        transformation_external_id=new_transformation.external_id, destination=f"my_{prefix}@email.com"
    )
    tn = cognite_client.transformations.notifications.create(notification)

    yield (tn, new_transformation.external_id)

    cognite_client.transformations.notifications.delete(id=tn.id)
    assert (
        len(
            cognite_client.transformations.notifications.list(transformation_external_id=new_transformation.external_id)
        )
        == 0
    )


class TestTransformationNotificationsAPI:
    def test_create(self, new_notification: TransformationNotification):
        assert (
            new_notification.destination == f"my_{prefix}@email.com"
            and new_notification.id is not None
            and new_notification.created_time is not None
            and new_notification.last_updated_time is not None
        )

    def test_create_by_external_id(self, new_notification_by_external_id: TransformationNotification):
        new_notification = new_notification_by_external_id[0]
        assert (
            new_notification.destination == f"my_{prefix}@email.com"
            and new_notification.id is not None
            and new_notification.created_time is not None
            and new_notification.last_updated_time is not None
        )

    def test_list_all(self, cognite_client, new_notification):
        retrieved_notifications = cognite_client.transformations.notifications.list()
        assert new_notification.id in [notification.id for notification in retrieved_notifications]

    def test_list_by_id(self, cognite_client, new_notification):
        retrieved_notifications = cognite_client.transformations.notifications.list(
            transformation_id=new_notification.transformation_id
        )
        assert new_notification.id in [notification.id for notification in retrieved_notifications]
        assert len(retrieved_notifications) == 1

    def test_list_by_external_id(self, cognite_client, new_notification_by_external_id):
        new_notification = new_notification_by_external_id[0]
        external_id = new_notification_by_external_id[1]
        retrieved_notifications = cognite_client.transformations.notifications.list(
            transformation_external_id=external_id
        )
        assert new_notification.id in [notification.id for notification in retrieved_notifications]
        assert len(retrieved_notifications) == 1

    def test_list_by_destination(self, cognite_client, new_notification):
        retrieved_notifications = cognite_client.transformations.notifications.list(
            destination=f"my_{prefix}@email.com"
        )
        assert new_notification.id in [notification.id for notification in retrieved_notifications]
        assert len(retrieved_notifications) == 1

    def test_notification_to_string(self, new_notification):
        dumped = str(new_notification)



import pytest

from cognite.client.data_classes import APIKey, Group, SecurityCategory, ServiceAccount, ServiceAccountList
from cognite.client.utils._auxiliary import random_string


class TestServiceAccountAPI:
    @pytest.mark.skip("ServiceAccount methods failing, probably due to api keys deprecation")
    def test_list(self, cognite_client):
        res = cognite_client.iam.service_accounts.list()
        assert isinstance(res, ServiceAccountList)
        assert len(res) > 0

    def test_create_and_delete(self, cognite_client):
        name = "test_sa_" + random_string(10)
        sa = cognite_client.iam.service_accounts.create(ServiceAccount(name=name))
        assert isinstance(sa, ServiceAccount)
        assert sa.id in {s.id for s in cognite_client.iam.service_accounts.list()}
        cognite_client.iam.service_accounts.delete(sa.id)
        assert sa.id not in {s.id for s in cognite_client.iam.service_accounts.list()}


@pytest.fixture(scope="module")
def service_account_id(cognite_client):
    service_accounts = cognite_client.iam.service_accounts.list()
    return next(sa for sa in service_accounts if sa.name == "admin").id


@pytest.mark.skip("most API keys methods failing, probably due to deprecation")
class TestAPIKeysAPI:
    def test_list_with_sa_id(self, cognite_client, service_account_id):
        res = cognite_client.iam.api_keys.list(service_account_id=service_account_id)
        assert len(res) > 0

    def test_list_all(self, cognite_client):
        res = cognite_client.iam.api_keys.list(all=True)
        assert len(res) > 0

    def test_list_deleted(self, cognite_client):
        res = cognite_client.iam.api_keys.list(include_deleted=True, all=True)
        assert len(res) > 0

    def test_create_and_delete(self, cognite_client, service_account_id):
        res = cognite_client.iam.api_keys.create(service_account_id)
        assert isinstance(res, APIKey)
        assert res.id in {k.id for k in cognite_client.iam.api_keys.list(all=True)}
        cognite_client.iam.api_keys.delete(res.id)
        assert res.id not in {k.id for k in cognite_client.iam.api_keys.list(all=True)}


@pytest.fixture(scope="module")
def group_id(cognite_client):
    return cognite_client.iam.groups.list()[0].id


class TestGroupsAPI:
    def test_list(self, cognite_client):
        res = cognite_client.iam.groups.list(all=True)
        assert len(res) > 0

    def test_create(self, cognite_client):
        group = cognite_client.iam.groups.create(
            Group(name="bla", capabilities=[{"eventsAcl": {"actions": ["READ"], "scope": {"all": {}}}}])
        )
        assert "bla" == group.name
        cognite_client.iam.groups.delete(group.id)


class TestSecurityCategoriesAPI:
    def test_list(self, cognite_client):
        res = cognite_client.iam.security_categories.list()
        assert len(res) > 0

    def test_create_and_delete(self, cognite_client):
        random_name = "test_" + random_string(10)
        res = cognite_client.iam.security_categories.create(SecurityCategory(name=random_name))
        assert res.id in {s.id for s in cognite_client.iam.security_categories.list()}
        cognite_client.iam.security_categories.delete(res.id)
        assert res.id not in {s.id for s in cognite_client.iam.security_categories.list()}



import json
from decimal import Decimal

import pytest

from cognite.client import utils
from cognite.client.exceptions import CogniteImportError


class TestCaseConversion:
    def test_to_camel_case(self):
        assert "camelCase" == utils._auxiliary.to_camel_case("camel_case")
        assert "camelCase" == utils._auxiliary.to_camel_case("camelCase")
        assert "a" == utils._auxiliary.to_camel_case("a")

    def test_to_snake_case(self):
        assert "snake_case" == utils._auxiliary.to_snake_case("snakeCase")
        assert "snake_case" == utils._auxiliary.to_snake_case("snake_case")
        assert "a" == utils._auxiliary.to_snake_case("a")


class TestLocalImport:
    @pytest.mark.dsl
    def test_local_import_single_ok(self):
        import pandas

        assert pandas == utils._auxiliary.local_import("pandas")

    @pytest.mark.dsl
    def test_local_import_multiple_ok(self):
        import numpy
        import pandas

        assert (pandas, numpy) == utils._auxiliary.local_import("pandas", "numpy")

    def test_local_import_single_fail(self):
        with pytest.raises(CogniteImportError, match="requires 'not-a-module' to be installed"):
            utils._auxiliary.local_import("not-a-module")

    @pytest.mark.dsl
    def test_local_import_multiple_fail(self):
        with pytest.raises(CogniteImportError, match="requires 'not-a-module' to be installed"):
            utils._auxiliary.local_import("pandas", "not-a-module")

    @pytest.mark.coredeps
    def test_dsl_deps_not_installed(self):
        for dep in ["numpy", "pandas"]:
            with pytest.raises(CogniteImportError, match=dep):
                utils._auxiliary.local_import(dep)


class TestUrlEncode:
    def test_url_encode(self):
        assert "/bla/yes%2Fno/bla" == utils._auxiliary.interpolate_and_url_encode("/bla/{}/bla", "yes/no")
        assert "/bla/123/bla/456" == utils._auxiliary.interpolate_and_url_encode("/bla/{}/bla/{}", "123", "456")


class TestJsonDumpDefault:
    def test_json_serializable_Decimal(self):
        with pytest.raises(TypeError):
            json.dumps(Decimal(1))

        assert json.dumps(Decimal(1), default=utils._auxiliary.json_dump_default)

    def test_json_not_serializable_sets(self):
        with pytest.raises(TypeError):
            json.dumps({1, 2})
        with pytest.raises(TypeError):
            json.dumps({1, 2})

    @pytest.mark.dsl
    def test_json_serializable_numpy(self):
        np = utils._auxiliary.local_import("numpy")
        arr = np.array([1.2, 3.4]).astype(np.float32)
        with pytest.raises(TypeError):
            json.dumps(arr)
        with pytest.raises(TypeError):
            json.dumps(arr[0])
        with pytest.raises(TypeError):  # core sdk makes it hard to serialize np.ndarray
            assert json.dumps(arr, default=utils._auxiliary.json_dump_default)
        assert json.dumps(arr[0], default=utils._auxiliary.json_dump_default)

    def test_json_serializable_object(self):
        class Obj:
            def __init__(self):
                self.x = 1

        with pytest.raises(TypeError):
            json.dumps(Obj())

        assert json.dumps({"x": 1}) == json.dumps(Obj(), default=utils._auxiliary.json_dump_default)

    @pytest.mark.dsl
    def test_json_serialiable_numpy_integer(self):
        import numpy as np

        inputs = [np.int32(1), np.int64(1)]
        for input in inputs:
            assert json.dumps(input, default=utils._auxiliary.json_dump_default)


class TestSplitIntoChunks:
    @pytest.mark.parametrize(
        "input, chunk_size, expected_output",
        [
            (["a", "b", "c"], 1, [["a"], ["b"], ["c"]]),
            (["a", "b", "c"], 2, [["a", "b"], ["c"]]),
            ([], 1000, []),
            (["a", "b", "c"], 3, [["a", "b", "c"]]),
            (["a", "b", "c"], 10, [["a", "b", "c"]]),
            ({"a": 1, "b": 2}, 1, [{"a": 1}, {"b": 2}]),
            ({"a": 1, "b": 2}, 2, [{"a": 1, "b": 2}]),
            ({"a": 1, "b": 2}, 3, [{"a": 1, "b": 2}]),
            ({}, 1, []),
        ],
    )
    def test_split_into_chunks(self, input, chunk_size, expected_output):
        actual_output = utils._auxiliary.split_into_chunks(input, chunk_size)
        assert len(actual_output) == len(expected_output)
        for element in expected_output:
            assert element in actual_output


class TestAssertions:
    @pytest.mark.parametrize("var, var_name, types", [(1, "var1", [int]), ("1", "var2", [int, str])])
    def test_assert_type_ok(self, var, var_name, types):
        utils._auxiliary.assert_type(var, var_name, types=types)

    @pytest.mark.parametrize("var, var_name, types", [("1", "var", [int, float]), ((1,), "var2", [dict, list])])
    def test_assert_type_fail(self, var, var_name, types):
        with pytest.raises(TypeError, match=str(types)):
            utils._auxiliary.assert_type(var, var_name, types)

    def test_assert_exactly_one_of_id_and_external_id(self):
        with pytest.raises(AssertionError):
            utils._auxiliary.assert_exactly_one_of_id_or_external_id(1, "1")
        utils._auxiliary.assert_exactly_one_of_id_or_external_id(1, None)
        utils._auxiliary.assert_exactly_one_of_id_or_external_id(None, "1")



from unittest.mock import MagicMock

import pytest

from cognite.client import CogniteClient
from cognite.client.testing import CogniteClientMock, monkeypatch_cognite_client
from tests.utils import set_env_var


def test_mock_cognite_client():
    with monkeypatch_cognite_client() as c_mock:
        c = CogniteClient()
        assert isinstance(c_mock, MagicMock)
        assert c_mock == c

        api_pairs = [
            (c.time_series, c_mock.time_series),
            (c.raw, c_mock.raw),
            (c.assets, c_mock.assets),
            (c.datapoints, c_mock.datapoints),
            (c.events, c_mock.events),
            (c.files, c_mock.files),
            (c.iam, c_mock.iam),
            (c.login, c_mock.login),
            (c.three_d, c_mock.three_d),
        ]
        for api, mock_api in api_pairs:
            assert isinstance(mock_api, MagicMock)
            assert api == mock_api

            with pytest.raises(AttributeError):
                api.does_not_exist


def test_cognite_client_accepts_arguments_during_and_after_mock():
    with monkeypatch_cognite_client():
        CogniteClient(api_key=None)
    with set_env_var("COGNITE_API_KEY", "BLA"):
        CogniteClient(api_key=None)


def test_client_mock_can_access_attributes_not_explicitly_defined_on_children():
    c_mock = CogniteClientMock()
    assert c_mock.config.max_workers



from cognite.client._cognite_client import CogniteClient as Client


class CogniteClient(Client):
    def __init__(self, *args, **kwargs):
        super().__init__(api_subversion="beta", *args, **kwargs)



import re

import pytest

from cognite.client.data_classes import Label, LabelDefinition, LabelDefinitionList
from tests.utils import jsgz_load


@pytest.fixture
def mock_labels_response(rsps, cognite_client):
    response_body = {
        "items": [
            {"name": "Pump", "description": "guess", "externalId": "PUMP", "createdTime": 1575892259245, "dataSetId": 1}
        ]
    }

    url_pattern = re.compile(re.escape(cognite_client.labels._get_base_url_with_base_path()) + "/.+")
    rsps.assert_all_requests_are_fired = False

    rsps.add(rsps.POST, url_pattern, status=200, json=response_body)
    rsps.add(rsps.GET, url_pattern, status=200, json=response_body)
    yield rsps


class TestLabels:
    def test_list(self, cognite_client, mock_labels_response):
        res = cognite_client.labels.list(external_id_prefix="P")
        assert "P" == jsgz_load(mock_labels_response.calls[0].request.body)["filter"]["externalIdPrefix"]
        assert mock_labels_response.calls[0].response.json()["items"] == res.dump(camel_case=True)

    def test_access_properties(self, cognite_client, mock_labels_response):
        res = cognite_client.labels.list(external_id_prefix="P")
        assert res[0].name == "Pump"
        assert res[0].description == "guess"
        assert res[0].external_id == "PUMP"
        assert res[0].created_time > 0

    def test_call(self, cognite_client, mock_labels_response):
        list(cognite_client.labels(limit=10))
        calls = mock_labels_response.calls
        assert 1 == len(calls)
        assert {"cursor": None, "limit": 10, "filter": {}} == jsgz_load(calls[0].request.body)

    def test_create_single(self, cognite_client, mock_labels_response):
        res = cognite_client.labels.create(LabelDefinition(external_id="1", name="my_label", description="text"))
        assert isinstance(res, LabelDefinition)
        assert mock_labels_response.calls[0].response.json()["items"][0] == res.dump(camel_case=True)
        assert {"items": [{"externalId": "1", "name": "my_label", "description": "text"}]} == jsgz_load(
            mock_labels_response.calls[0].request.body
        )

    def test_create_multiple(self, cognite_client, mock_labels_response):
        res = cognite_client.labels.create([LabelDefinition(external_id="1"), LabelDefinition(external_id="2")])
        assert isinstance(res, LabelDefinitionList)
        assert mock_labels_response.calls[0].response.json()["items"] == res.dump(camel_case=True)
        assert {"items": [{"externalId": "1"}, {"externalId": "2"}]} == jsgz_load(
            mock_labels_response.calls[0].request.body
        )

    def test_delete_single(self, cognite_client, mock_labels_response):
        res = cognite_client.labels.delete(external_id="PUMP")
        assert {"items": [{"externalId": "PUMP"}]} == jsgz_load(mock_labels_response.calls[0].request.body)
        assert res is None

    def test_delete_multiple(self, cognite_client, mock_labels_response):
        res = cognite_client.labels.delete(external_id=["PUMP", "VALVE"])
        assert {"items": [{"externalId": "PUMP"}, {"externalId": "VALVE"}]} == jsgz_load(
            mock_labels_response.calls[0].request.body
        )
        assert res is None

    def test_create_labels_using_wrong_type(self, cognite_client):
        with pytest.raises(TypeError):
            cognite_client.labels.create(Label(external_id="1", name="my_label"))
        with pytest.raises(TypeError):
            cognite_client.labels.create([Label(external_id="1", name="my_label")])

    def test_load_list(self):
        assert Label._load_list(None) == None
        labels = [{"externalId": "a"}, "b", Label("c"), LabelDefinition("d")]
        assert Label._load_list(labels) == [Label("a"), Label("b"), Label("c"), Label("d")]

    def test_list_with_dataset_ids(self, cognite_client, mock_labels_response):
        res = cognite_client.labels.list(data_set_ids=[123], data_set_external_ids=["x"])
        assert res[0].data_set_id == 1
        assert [{"id": 123}, {"externalId": "x"}] == jsgz_load(mock_labels_response.calls[0].request.body)["filter"][
            "dataSetIds"
        ]



import pytest

from cognite.client.data_classes import ThreeDModelRevisionUpdate, ThreeDModelUpdate


@pytest.fixture(scope="class")
def test_revision(cognite_client):
    model_id = cognite_client.three_d.models.list(published=True, limit=1)[0].id
    revision = cognite_client.three_d.revisions.list(model_id=model_id, limit=1)[0]
    return revision, model_id


@pytest.fixture(scope="class")
def new_model(cognite_client):
    res = cognite_client.three_d.models.create(name="NewTestModel")
    yield res
    cognite_client.three_d.models.delete(id=res.id)
    assert cognite_client.three_d.models.retrieve(id=res.id) is None


@pytest.fixture(scope="class")
def test_nodes_tree_index_order(cognite_client):
    revision, model_id = test_revision
    return cognite_client.three_d.list_nodes(model_id=model_id, revision_id=revision.id)


class TestThreeDModelsAPI:
    def test_list_and_retrieve(self, cognite_client):
        res = cognite_client.three_d.models.list(limit=1)
        assert 1 == len(res)
        res = [r for r in cognite_client.three_d.models(limit=None) if r.name == "MyModel775"][0]
        assert res == cognite_client.three_d.models.retrieve(res.id)

    def test_update_with_resource(self, new_model, cognite_client):
        model = new_model
        model.name = "NewTestModelVer2"
        res = cognite_client.three_d.models.update(model)
        assert model.name == res.name

    def test_partial_update(self, new_model, cognite_client):
        added_metadata = {"key": "value"}
        update = ThreeDModelUpdate(id=new_model.id).metadata.add(added_metadata)
        res = cognite_client.three_d.models.update(update)
        assert added_metadata == res.metadata


class TestThreeDRevisionsAPI:
    @pytest.mark.skip(reason="missing a 3d model to test revision against")
    def test_list_and_retrieve(self, test_revision, cognite_client):
        revision, model_id = test_revision
        assert revision == cognite_client.three_d.revisions.retrieve(model_id=model_id, id=revision.id)
        res = cognite_client.three_d.revisions.list(model_id=model_id)
        assert len(res) > 0
        res = cognite_client.three_d.revisions.list_nodes(model_id=model_id, revision_id=revision.id)
        assert len(res) > 0

    @pytest.mark.skip(reason="missing a 3d model to test revision against")
    def test_list_ancestor_nodes(self, test_revision, cognite_client):
        revision, model_id = test_revision
        node_id = test_nodes_tree_index_order[0].id
        res = cognite_client.three_d.revisions.list_ancestor_nodes(
            model_id=model_id, revision_id=revision.id, node_id=node_id
        )
        assert len(res) > 0

    @pytest.mark.skip(reason="missing a 3d model to test revision against")
    def test_list_nodes_partitions(self, test_revision, cognite_client):
        revision, model_id = test_revision
        nodes = cognite_client.three_d.revisions.list_nodes(
            model_id=model_id, revision_id=revision.id, partitions=4, sort_by_node_id=True
        )
        assert len(nodes) == len(test_nodes_tree_index_order)
        assert test_node_tree_index_order[nodes[0].tree_index].id == nodes[0].id

    @pytest.mark.skip(reason="missing a 3d model to test revision against")
    def test_update_with_resource(self, cognite_client, test_revision):
        revision, model_id = test_revision
        revision.metadata = {"key": "value"}
        cognite_client.three_d.revisions.update(model_id, revision)

    @pytest.mark.skip(reason="missing a 3d model to test revision against")
    def test_partial_update(self, cognite_client, test_revision):
        revision, model_id = test_revision
        added_metadata = {"key": "value"}
        revision_update = ThreeDModelRevisionUpdate(id=revision.id).metadata.add(added_metadata)
        cognite_client.three_d.revisions.update(model_id=model_id, item=revision_update)


class TestThreeDFilesAPI:
    @pytest.mark.skip(reason="missing a 3d model to test revision against")
    def test_retrieve(self, cognite_client, test_revision):
        revision, model_id = test_revision
        project = cognite_client._config.project
        url = "/api/v1/projects/{}/3d/reveal/models/{}/revisions/{}".format(project, model_id, revision.id)
        response = cognite_client.get(url=url)
        threedFileId = response.json()["sceneThreedFiles"][0]["fileId"]

        res = cognite_client.three_d.files.retrieve(threedFileId)
        assert len(res) > 0



import json
import os
import re
from io import BufferedReader, TextIOWrapper
from tempfile import TemporaryDirectory

import pytest

from cognite.client._api.files import FileMetadata, FileMetadataList, FileMetadataUpdate
from cognite.client.data_classes import (
    FileMetadataFilter,
    GeoLocation,
    GeoLocationFilter,
    Geometry,
    GeometryFilter,
    Label,
    LabelFilter,
    TimestampRange,
)
from cognite.client.exceptions import CogniteAPIError
from tests.utils import jsgz_load, set_request_limit


@pytest.fixture
def mock_geo_location():
    geometry = Geometry(type="Point", coordinates=[35, 10])
    yield GeoLocation(type="Feature", geometry=geometry)


@pytest.fixture
def mock_files_response(rsps, cognite_client, mock_geo_location):
    response_body = {
        "items": [
            {
                "externalId": "string",
                "name": "string",
                "source": "string",
                "mimeType": "string",
                "metadata": {"metadata-key": "metadata-value"},
                "assetIds": [1],
                "labels": [{"externalId": "WELL LOG"}],
                "geoLocation": mock_geo_location,
                "id": 1,
                "uploaded": True,
                "uploadedTime": 0,
                "createdTime": 0,
                "lastUpdatedTime": 0,
            }
        ]
    }

    url_pattern = re.compile(re.escape(cognite_client.files._get_base_url_with_base_path()) + "/.+")
    rsps.assert_all_requests_are_fired = False

    rsps.add(rsps.POST, url_pattern, status=200, json=response_body)
    rsps.add(rsps.GET, url_pattern, status=200, json=response_body)
    yield rsps


@pytest.fixture
def mock_file_upload_response(rsps, cognite_client, mock_geo_location):
    response_body = {
        "externalId": "string",
        "name": "string",
        "source": "string",
        "mimeType": "string",
        "metadata": {},
        "assetIds": [1],
        "labels": [{"externalId": "WELL LOG"}],
        "geoLocation": mock_geo_location,
        "id": 1,
        "uploaded": True,
        "uploadedTime": 0,
        "createdTime": 0,
        "lastUpdatedTime": 0,
        "uploadUrl": "https://upload.here",
    }
    rsps.add(rsps.POST, cognite_client.files._get_base_url_with_base_path() + "/files", status=200, json=response_body)
    rsps.add(rsps.PUT, "https://upload.here", status=200)
    yield rsps


@pytest.fixture
def mock_file_create_response(rsps, cognite_client, mock_geo_location):
    response_body = {
        "externalId": "string",
        "name": "string",
        "source": "string",
        "mimeType": "string",
        "metadata": {},
        "assetIds": [1],
        "labels": [{"externalId": "WELL LOG"}],
        "geoLocation": mock_geo_location,
        "id": 1,
        "uploaded": False,
        "uploadedTime": 0,
        "createdTime": 0,
        "lastUpdatedTime": 0,
        "uploadUrl": "https://upload.here",
    }
    rsps.add(rsps.POST, cognite_client.files._get_base_url_with_base_path() + "/files", status=200, json=response_body)
    yield rsps


@pytest.fixture
def mock_file_download_response(rsps, cognite_client):
    rsps.add(
        rsps.POST,
        cognite_client.files._get_base_url_with_base_path() + "/files/byids",
        status=200,
        json={"items": [{"id": 1, "name": "file1"}, {"externalId": "2", "name": "file2"}]},
    )

    def download_link_callback(request):
        identifier = jsgz_load(request.body)["items"][0]
        response = {}
        if "id" in identifier:
            response = {"items": [{"id": 1, "downloadUrl": "https://download.file1.here"}]}
        elif "externalId" in identifier:
            response = {"items": [{"externalId": "2", "downloadUrl": "https://download.file2.here"}]}
        return 200, {}, json.dumps(response)

    rsps.add_callback(
        rsps.POST,
        cognite_client.files._get_base_url_with_base_path() + "/files/downloadlink",
        callback=download_link_callback,
        content_type="application/json",
    )
    rsps.add(rsps.GET, "https://download.file1.here", status=200, body="content1")
    rsps.add(rsps.GET, "https://download.file2.here", status=200, body="content2")
    yield rsps


@pytest.fixture
def mock_file_download_response_one_fails(rsps, cognite_client):
    rsps.add(
        rsps.POST,
        cognite_client.files._get_base_url_with_base_path() + "/files/byids",
        status=200,
        json={
            "items": [
                {"id": 1, "externalId": "success", "name": "file1"},
                {"externalId": "fail", "id": 2, "name": "file2"},
            ]
        },
    )

    def download_link_callback(request):
        identifier = jsgz_load(request.body)["items"][0]
        if "id" in identifier:
            return 200, {}, json.dumps({"items": [{"id": 1, "downloadUrl": "https://download.file1.here"}]})
        elif "externalId" in identifier:
            return (400, {}, json.dumps({"error": {"message": "User error", "code": 400}}))

    rsps.add_callback(
        rsps.POST,
        cognite_client.files._get_base_url_with_base_path() + "/files/downloadlink",
        callback=download_link_callback,
        content_type="application/json",
    )
    rsps.add(rsps.GET, "https://download.file1.here", status=200, body="content1")
    yield rsps


class TestFilesAPI:
    def test_create(self, cognite_client, mock_file_create_response):
        file_metadata = FileMetadata(name="bla")
        returned_file_metadata, upload_url = cognite_client.files.create(file_metadata)
        response_body = mock_file_create_response.calls[0].response.json()
        assert FileMetadata._load(response_body) == returned_file_metadata
        assert response_body["uploadUrl"] == upload_url

    def test_create_with_label(self, cognite_client, mock_file_create_response):
        file_metadata = FileMetadata(name="bla", labels=[Label(external_id="WELL LOG")])
        returned_file_metadata, upload_url = cognite_client.files.create(file_metadata)
        response_body = mock_file_create_response.calls[0].response.json()
        assert FileMetadata._load(response_body) == returned_file_metadata
        assert response_body["uploadUrl"] == upload_url
        assert response_body["labels"][0]["externalId"] == "WELL LOG"

    def test_create_with_label_request(self, cognite_client, mock_file_create_response):
        file_metadata = FileMetadata(name="bla", labels=[Label(external_id="WELL LOG")])
        returned_file_metadata, upload_url = cognite_client.files.create(file_metadata)
        response_body = mock_file_create_response.calls[0].response.json()
        request_body = jsgz_load(mock_file_create_response.calls[0].request.body)
        assert FileMetadata._load(response_body) == returned_file_metadata
        assert all(body["labels"][0]["externalId"] == "WELL LOG" for body in [request_body, response_body])

    def test_create_with_geoLocation(self, cognite_client, mock_file_create_response, mock_geo_location):
        file_metadata = FileMetadata(name="bla", geo_location=mock_geo_location)
        returned_file_metadata, upload_url = cognite_client.files.create(file_metadata)
        response_body = mock_file_create_response.calls[0].response.json()
        assert FileMetadata._load(response_body) == returned_file_metadata
        assert response_body["geoLocation"] == mock_geo_location

    def test_create_geoLocation_with_invalid_geometry_type(self):
        with pytest.raises(ValueError):
            _ = Geometry(type="someInvalidType", coordinates=[1, 2])

    def test_create_geoLocation_with_invalid_geojson_type(self):
        g = Geometry(type="Point", coordinates=[1, 2])
        with pytest.raises(ValueError):
            _ = GeoLocation(type="FeatureCollection", geometry=g)

    def test_create_with_geoLocation_request(self, cognite_client, mock_file_create_response, mock_geo_location):
        file_metadata = FileMetadata(name="bla", geo_location=mock_geo_location)
        returned_file_metadata, upload_url = cognite_client.files.create(file_metadata)
        response_body = mock_file_create_response.calls[0].response.json()
        request_body = jsgz_load(mock_file_create_response.calls[0].request.body)
        assert FileMetadata._load(response_body) == returned_file_metadata
        assert all(body["geoLocation"] == mock_geo_location for body in [request_body, response_body])

    def test_retrieve_single(self, cognite_client, mock_files_response):
        res = cognite_client.files.retrieve(id=1)
        assert isinstance(res, FileMetadata)
        assert mock_files_response.calls[0].response.json()["items"][0] == res.dump(camel_case=True)

    def test_retrieve_multiple(self, cognite_client, mock_files_response):
        res = cognite_client.files.retrieve_multiple(ids=[1])
        assert isinstance(res, FileMetadataList)
        assert mock_files_response.calls[0].response.json()["items"] == res.dump(camel_case=True)

    def test_list(self, cognite_client, mock_files_response):
        res = cognite_client.files.list(source="bla", limit=10)
        assert isinstance(res, FileMetadataList)
        assert mock_files_response.calls[0].response.json()["items"] == res.dump(camel_case=True)
        assert "bla" == jsgz_load(mock_files_response.calls[0].request.body)["filter"]["source"]
        assert 10 == jsgz_load(mock_files_response.calls[0].request.body)["limit"]

    def test_list_params(self, cognite_client, mock_files_response):
        cognite_client.files.list(
            root_asset_ids=[1, 2], root_asset_external_ids=["a"], data_set_external_ids=["x"], limit=10
        )
        calls = mock_files_response.calls
        assert 1 == len(calls)
        assert {
            "cursor": None,
            "limit": 10,
            "filter": {
                "rootAssetIds": [{"id": 1}, {"id": 2}, {"externalId": "a"}],
                "dataSetIds": [{"externalId": "x"}],
            },
        } == jsgz_load(calls[0].request.body)

    def test_list_subtrees(self, cognite_client, mock_files_response):
        cognite_client.files.list(asset_subtree_ids=[1], asset_subtree_external_ids=["a"], limit=10)
        calls = mock_files_response.calls
        assert 1 == len(calls)
        assert {
            "cursor": None,
            "limit": 10,
            "filter": {"assetSubtreeIds": [{"id": 1}, {"externalId": "a"}]},
        } == jsgz_load(calls[0].request.body)

    def test_filter_directory(self, cognite_client, mock_files_response):
        cognite_client.files.list(directory_prefix="/test", limit=10)
        calls = mock_files_response.calls
        assert len(calls) == 1
        assert jsgz_load(calls[0].request.body) == {"cursor": None, "filter": {"directoryPrefix": "/test"}, "limit": 10}

    def test_filter_geoLocation(self, cognite_client, mock_files_response):
        cognite_client.files.list(
            geo_location=GeoLocationFilter(relation="within", shape=GeometryFilter(type="Point", coordinates=[35, 10])),
            limit=10,
        )
        calls = mock_files_response.calls
        assert len(calls) == 1
        assert jsgz_load(calls[0].request.body) == {
            "cursor": None,
            "filter": {"geoLocation": {"relation": "within", "shape": {"type": "Point", "coordinates": [35, 10]}}},
            "limit": 10,
        }

    def test_list_with_time_dict(self, cognite_client, mock_files_response):
        cognite_client.files.list(created_time={"min": 20})
        assert 20 == jsgz_load(mock_files_response.calls[0].request.body)["filter"]["createdTime"]["min"]
        assert "max" not in jsgz_load(mock_files_response.calls[0].request.body)["filter"]["createdTime"]

    def test_list_with_timestamp_range(self, cognite_client, mock_files_response):
        cognite_client.files.list(created_time=TimestampRange(min=20))
        assert 20 == jsgz_load(mock_files_response.calls[0].request.body)["filter"]["createdTime"]["min"]
        assert "max" not in jsgz_load(mock_files_response.calls[0].request.body)["filter"]["createdTime"]

    def test_delete_single(self, cognite_client, mock_files_response):
        res = cognite_client.files.delete(id=1)
        assert {"items": [{"id": 1}]} == jsgz_load(mock_files_response.calls[0].request.body)
        assert res is None

    def test_delete_multiple(self, cognite_client, mock_files_response):
        res = cognite_client.files.delete(id=[1])
        assert {"items": [{"id": 1}]} == jsgz_load(mock_files_response.calls[0].request.body)
        assert res is None

    def test_update_with_resource_class(self, cognite_client, mock_files_response):
        res = cognite_client.files.update(FileMetadata(id=1, source="bla"))
        assert isinstance(res, FileMetadata)
        assert {"items": [{"id": 1, "update": {"source": {"set": "bla"}}}]} == jsgz_load(
            mock_files_response.calls[0].request.body
        )

    def test_update_with_update_class(self, cognite_client, mock_files_response):
        res = cognite_client.files.update(FileMetadataUpdate(id=1).source.set("bla"))
        assert isinstance(res, FileMetadata)
        assert {"items": [{"id": 1, "update": {"source": {"set": "bla"}}}]} == jsgz_load(
            mock_files_response.calls[0].request.body
        )

    def test_update_labels_single(self, cognite_client, mock_files_response):
        cognite_client.files.update([FileMetadataUpdate(id=1).labels.add("PUMP").labels.remove("WELL LOG")])
        expected = {"labels": {"add": [{"externalId": "PUMP"}], "remove": [{"externalId": "WELL LOG"}]}}
        assert jsgz_load(mock_files_response.calls[0].request.body)["items"][0]["update"] == expected

    def test_update_labels_multiple(self, cognite_client, mock_files_response):
        cognite_client.files.update(
            [FileMetadataUpdate(id=1).labels.add(["PUMP", "ROTATING_EQUIPMENT"]).labels.remove(["WELL LOG"])]
        )
        expected = {
            "labels": {
                "add": [{"externalId": "PUMP"}, {"externalId": "ROTATING_EQUIPMENT"}],
                "remove": [{"externalId": "WELL LOG"}],
            }
        }
        assert jsgz_load(mock_files_response.calls[0].request.body)["items"][0]["update"] == expected

    # resource.update doesn't support full replacement of labels (set operation)
    def test_ignore_labels_resource_class(self, cognite_client, mock_files_response):
        cognite_client.files.update(FileMetadata(id=1, labels=[Label(external_id="Pump")], external_id="newId"))
        assert jsgz_load(mock_files_response.calls[0].request.body)["items"][0]["update"] == {
            "externalId": {"set": "newId"}
        }

    def test_labels_filter_contains_all(self, cognite_client, mock_files_response):
        my_label_filter = LabelFilter(contains_all=["WELL LOG", "VERIFIED"])
        cognite_client.files.list(labels=my_label_filter)
        assert jsgz_load(mock_files_response.calls[0].request.body)["filter"]["labels"] == {
            "containsAll": [{"externalId": "WELL LOG"}, {"externalId": "VERIFIED"}]
        }

    def test_labels_filter_contains_any(self, cognite_client, mock_files_response):
        my_label_filter = LabelFilter(contains_any=["WELL LOG", "WELL REPORT"])
        cognite_client.files.list(labels=my_label_filter)
        assert jsgz_load(mock_files_response.calls[0].request.body)["filter"]["labels"] == {
            "containsAny": [{"externalId": "WELL LOG"}, {"externalId": "WELL REPORT"}]
        }

    def test_update_multiple(self, cognite_client, mock_files_response):
        res = cognite_client.files.update(
            [FileMetadataUpdate(id=1).source.set(None), FileMetadata(external_id="2", source="bla")]
        )
        assert isinstance(res, FileMetadataList)
        assert {
            "items": [
                {"id": 1, "update": {"source": {"setNull": True}}},
                {"externalId": "2", "update": {"source": {"set": "bla"}}},
            ]
        } == jsgz_load(mock_files_response.calls[0].request.body)

    def test_iter_single(self, cognite_client, mock_files_response):
        for file in cognite_client.files:
            assert isinstance(file, FileMetadata)
            assert mock_files_response.calls[0].response.json()["items"][0] == file.dump(camel_case=True)

    def test_iter_chunk(self, cognite_client, mock_files_response):
        for file in cognite_client.files(chunk_size=1):
            assert isinstance(file, FileMetadataList)
            assert mock_files_response.calls[0].response.json()["items"] == file.dump(camel_case=True)

    def test_search(self, cognite_client, mock_files_response):
        res = cognite_client.files.search(filter=FileMetadataFilter(external_id_prefix="abc"))
        assert mock_files_response.calls[0].response.json()["items"] == res.dump(camel_case=True)
        assert {"search": {"name": None}, "filter": {"externalIdPrefix": "abc"}, "limit": 100} == jsgz_load(
            mock_files_response.calls[0].request.body
        )

    @pytest.mark.parametrize("filter_field", ["external_id_prefix", "externalIdPrefix"])
    def test_search_dict_filter(self, cognite_client, mock_files_response, filter_field):
        res = cognite_client.files.search(filter={filter_field: "abc"})
        assert mock_files_response.calls[0].response.json()["items"] == res.dump(camel_case=True)
        assert {"search": {"name": None}, "filter": {"externalIdPrefix": "abc"}, "limit": 100} == jsgz_load(
            mock_files_response.calls[0].request.body
        )

    def test_upload(self, cognite_client, mock_file_upload_response):
        dir = os.path.join(os.path.dirname(__file__), "files_for_test_upload")
        path = os.path.join(dir, "file_for_test_upload_1.txt")
        res = cognite_client.files.upload(path, name="bla", directory=dir)
        response_body = mock_file_upload_response.calls[0].response.json()
        del response_body["uploadUrl"]
        assert FileMetadata._load(response_body) == res
        assert "https://upload.here/" == mock_file_upload_response.calls[1].request.url
        assert {"name": "bla", "directory": dir} == jsgz_load(mock_file_upload_response.calls[0].request.body)
        assert isinstance(mock_file_upload_response.calls[1].request.body, BufferedReader)

    def test_upload_with_external_id(self, cognite_client, mock_file_upload_response):
        path = os.path.join(os.path.dirname(__file__), "files_for_test_upload", "file_for_test_upload_1.txt")
        cognite_client.files.upload(path, external_id="blabla", name="bla", data_set_id=42)

    def test_upload_no_name(self, cognite_client, mock_file_upload_response):
        dir = os.path.join(os.path.dirname(__file__), "files_for_test_upload")
        path = os.path.join(dir, "file_for_test_upload_1.txt")
        cognite_client.files.upload(path, directory=dir)
        assert {"name": "file_for_test_upload_1.txt", "directory": dir} == jsgz_load(
            mock_file_upload_response.calls[0].request.body
        )

    def test_upload_set_directory(self, cognite_client, mock_file_upload_response):
        set_dir = "/Some/custom/directory"
        dir = os.path.join(os.path.dirname(__file__), "files_for_test_upload")
        path = os.path.join(dir, "file_for_test_upload_1.txt")
        cognite_client.files.upload(path, directory=set_dir)
        assert {"name": "file_for_test_upload_1.txt", "directory": set_dir} == jsgz_load(
            mock_file_upload_response.calls[0].request.body
        )

    def test_upload_from_directory(self, cognite_client, mock_file_upload_response):
        path = os.path.join(os.path.dirname(__file__), "files_for_test_upload")
        res = cognite_client.files.upload(path=path, asset_ids=[1, 2])
        response_body = mock_file_upload_response.calls[0].response.json()
        del response_body["uploadUrl"]
        assert FileMetadataList([FileMetadata._load(response_body), FileMetadata._load(response_body)]) == res
        assert 4 == len(mock_file_upload_response.calls)
        for call in mock_file_upload_response.calls:
            payload = call.request.body
            if isinstance(payload, BufferedReader):
                continue
            else:
                json = jsgz_load(payload)
                assert [1, 2] == json["assetIds"]
                assert json["name"] in ["file_for_test_upload_1.txt", "file_for_test_upload_2.txt"]

    def test_upload_from_directory_fails(self, cognite_client, rsps):
        rsps.add(rsps.POST, cognite_client.files._get_base_url_with_base_path() + "/files", status=400, json={})
        path = os.path.join(os.path.dirname(__file__), "files_for_test_upload")
        with pytest.raises(CogniteAPIError) as e:
            cognite_client.files.upload(path=path)
        assert "file_for_test_upload_1.txt" in e.value.failed
        assert "file_for_test_upload_2.txt" in e.value.failed

    def test_upload_from_directory_recursively(self, cognite_client, mock_file_upload_response):
        path = os.path.join(os.path.dirname(__file__), "files_for_test_upload")
        res = cognite_client.files.upload(path=path, recursive=True, asset_ids=[1, 2])
        response_body = mock_file_upload_response.calls[0].response.json()
        del response_body["uploadUrl"]
        assert FileMetadataList([FileMetadata._load(response_body) for _ in range(3)]) == res
        assert 6 == len(mock_file_upload_response.calls)
        for call in mock_file_upload_response.calls:
            payload = call.request.body
            if isinstance(payload, BufferedReader):
                continue
            else:
                json = jsgz_load(payload)
                assert json["name"] in [
                    "file_for_test_upload_1.txt",
                    "file_for_test_upload_2.txt",
                    "file_for_test_upload_3.txt",
                ]
            assert [1, 2] == json["assetIds"]

    def test_upload_from_memory(self, cognite_client, mock_file_upload_response):
        res = cognite_client.files.upload_bytes(content=b"content", name="bla")
        response_body = mock_file_upload_response.calls[0].response.json()
        del response_body["uploadUrl"]
        assert FileMetadata._load(response_body) == res
        assert "https://upload.here/" == mock_file_upload_response.calls[1].request.url
        assert {"name": "bla"} == jsgz_load(mock_file_upload_response.calls[0].request.body)
        assert b"content" == mock_file_upload_response.calls[1].request.body

    def test_upload_using_file_handle(self, cognite_client, mock_file_upload_response):
        path = os.path.join(os.path.dirname(__file__), "files_for_test_upload", "file_for_test_upload_1.txt")
        with open(path) as fh:
            res = cognite_client.files.upload_bytes(fh, name="bla")
        response_body = mock_file_upload_response.calls[0].response.json()
        del response_body["uploadUrl"]
        assert FileMetadata._load(response_body) == res
        assert "https://upload.here/" == mock_file_upload_response.calls[1].request.url
        assert {"name": "bla"} == jsgz_load(mock_file_upload_response.calls[0].request.body)
        assert isinstance(mock_file_upload_response.calls[1].request.body, TextIOWrapper)

    def test_upload_path_does_not_exist(self, cognite_client):
        with pytest.raises(ValueError, match="does not exist"):
            cognite_client.files.upload(path="/no/such/path")

    def test_download(self, cognite_client, mock_file_download_response):
        with TemporaryDirectory() as dir:
            res = cognite_client.files.download(directory=dir, id=[1], external_id=["2"])
            assert {"items": [{"id": 1}, {"externalId": "2"}]} == jsgz_load(
                mock_file_download_response.calls[0].request.body
            )
            assert res is None
            fp1 = os.path.join(dir, "file1")
            fp2 = os.path.join(dir, "file2")
            assert os.path.isfile(fp1)
            assert os.path.isfile(fp2)
            with open(fp1, "rb") as fh:
                assert b"content1" == fh.read()
            with open(fp2, "rb") as fh:
                assert b"content2" == fh.read()

    @pytest.fixture
    def mock_byids_response__file_with_double_dots(self, rsps, cognite_client):
        filename = "../file1"
        rsps.add(
            rsps.POST,
            cognite_client.files._get_base_url_with_base_path() + "/files/byids",
            status=200,
            json={"items": [{"id": 1, "name": filename}]},
        )
        yield rsps

    def test_download_file_outside_download_directory(self, cognite_client, mock_byids_response__file_with_double_dots):
        with TemporaryDirectory() as dir:
            with pytest.raises(RuntimeError, match="not inside download directory"):
                cognite_client.files.download(directory=dir, id=[1])

    def test_download_one_file_fails(self, cognite_client, mock_file_download_response_one_fails):
        with TemporaryDirectory() as dir:
            with pytest.raises(CogniteAPIError) as e:
                cognite_client.files.download(directory=dir, id=[1], external_id="fail")
            assert [FileMetadata(id=1, name="file1", external_id="success")] == e.value.successful
            assert [FileMetadata(id=2, name="file2", external_id="fail")] == e.value.failed
            assert os.path.isfile(os.path.join(dir, "file1"))

    def test_download_file_to_path(self, cognite_client, mock_file_download_response):
        mock_file_download_response.assert_all_requests_are_fired = False
        with TemporaryDirectory() as dir:
            file_path = os.path.join(dir, "my_downloaded_file.txt")
            res = cognite_client.files.download_to_path(path=file_path, id=1)
            assert res is None
            assert os.path.isfile(file_path)
            with open(file_path, "rb") as f:
                assert b"content1" == f.read()

    def test_download_to_memory(self, cognite_client, mock_file_download_response):
        mock_file_download_response.assert_all_requests_are_fired = False
        res = cognite_client.files.download_bytes(id=1)
        assert {"items": [{"id": 1}]} == jsgz_load(mock_file_download_response.calls[0].request.body)
        assert res == b"content1"

    def test_download_ids_over_limit(self, cognite_client, mock_file_download_response):
        with set_request_limit(cognite_client.files, 1):
            with TemporaryDirectory() as dir:
                res = cognite_client.files.download(directory=dir, id=[1], external_id=["2"])
                bodies = [jsgz_load(mock_file_download_response.calls[i].request.body) for i in range(2)]
                assert {"items": [{"id": 1}]} in bodies
                assert {"items": [{"externalId": "2"}]} in bodies
                assert res is None
                assert os.path.isfile(os.path.join(dir, "file1"))
                assert os.path.isfile(os.path.join(dir, "file2"))

    def test_files_update_object(self):
        assert isinstance(
            FileMetadataUpdate(1)
            .asset_ids.add([])
            .asset_ids.remove([])
            .external_id.set("1")
            .external_id.set(None)
            .directory.set("/some/new/directory")
            .metadata.add({})
            .metadata.remove([])
            .labels.add(["WELL LOG"])
            .labels.remove(["CV"])
            .geoLocation.set(mock_geo_location)
            .geoLocation.set(None)
            .source.set(1)
            .source.set(None),
            FileMetadataUpdate,
        )


@pytest.fixture
def mock_files_empty(rsps, cognite_client):
    url_pattern = re.compile(re.escape(cognite_client.files._get_base_url_with_base_path()) + "/.+")
    rsps.add(rsps.POST, url_pattern, status=200, json={"items": []})
    yield rsps


@pytest.mark.dsl
class TestPandasIntegration:
    def test_file_list_to_pandas(self, cognite_client, mock_files_response):
        import pandas as pd

        df = cognite_client.files.list().to_pandas()
        assert isinstance(df, pd.DataFrame)
        assert 1 == df.shape[0]
        assert {"metadata-key": "metadata-value"} == df["metadata"][0]

    def test_file_list_to_pandas_empty(self, cognite_client, mock_files_empty):
        import pandas as pd

        df = cognite_client.files.list().to_pandas()
        assert isinstance(df, pd.DataFrame)
        assert df.empty

    def test_file_to_pandas(self, cognite_client, mock_files_response):
        import pandas as pd

        df = cognite_client.files.retrieve(id=1).to_pandas()
        assert isinstance(df, pd.DataFrame)
        assert "metadata" not in df.columns
        assert [1] == df.loc["assetIds"][0]
        assert "metadata-value" == df.loc["metadata-key"][0]



import os
import pprint
import sys
from typing import *

from cognite.client import utils
from cognite.client._version import __api_subversion__
from cognite.client.exceptions import CogniteAPIKeyError

_DEFAULT_API_SUBVERSION = __api_subversion__


class _ThreadLocalConfig:
    def __init__(self):
        self.api_key = None
        self.project = None
        if "cognite._thread_local" in sys.modules:
            from cognite._thread_local import credentials

            thread_local_api_key = getattr(credentials, "api_key", None)
            thread_local_project = getattr(credentials, "project", None)
            self.api_key = thread_local_api_key
            self.project = thread_local_project


class _DefaultConfig:
    def __init__(self):
        thread_local = _ThreadLocalConfig()

        # Per client
        self.api_key = thread_local.api_key or os.getenv("COGNITE_API_KEY")
        self.api_subversion = os.getenv("COGNITE_API_VERSION") or _DEFAULT_API_SUBVERSION
        self.project = thread_local.project or os.getenv("COGNITE_PROJECT")
        self.client_name = os.getenv("COGNITE_CLIENT_NAME")
        self.base_url = os.getenv("COGNITE_BASE_URL", "https://api.cognitedata.com")
        self.max_workers = int(os.getenv("COGNITE_MAX_WORKERS", 10))
        self.headers = {}
        self.timeout = int(os.getenv("COGNITE_TIMEOUT", 30))
        self.file_transfer_timeout = int(os.getenv("COGNITE_FILE_TRANSFER_TIMEOUT", 600))
        self.token_client_id = os.getenv("COGNITE_CLIENT_ID")
        self.token_client_secret = os.getenv("COGNITE_CLIENT_SECRET")
        self.token_url = os.getenv("COGNITE_TOKEN_URL")
        self.token_scopes = os.getenv("COGNITE_TOKEN_SCOPES", "").split(",")
        self.token_custom_args = {}

        # Global
        self.disable_gzip = os.getenv("COGNITE_DISABLE_GZIP", False)
        self.disable_pypi_version_check = os.getenv("COGNITE_DISABLE_PYPI_VERSION_CHECK", False)
        self.status_forcelist = self._get_status_forcelist()
        self.max_retries = int(os.getenv("COGNITE_MAX_RETRIES", 10))
        self.max_retry_backoff = int(os.getenv("COGNITE_MAX_RETRY_BACKOFF", 30))
        self.max_connection_pool_size = int(os.getenv("COGNITE_MAX_CONNECTION_POOL_SIZE", 50))
        self.disable_ssl = os.getenv("COGNITE_DISABLE_SSL", False)

    @staticmethod
    def _get_status_forcelist():
        env_forcelist = os.getenv("COGNITE_STATUS_FORCELIST")
        if env_forcelist is None:
            return [429, 502, 503, 504]
        return [int(c) for c in env_forcelist.split(",")]


class ClientConfig(_DefaultConfig):
    def __init__(
        self,
        api_key: Optional[str] = None,
        api_subversion: Optional[str] = None,
        project: Optional[str] = None,
        client_name: Optional[str] = None,
        base_url: Optional[str] = None,
        max_workers: Optional[int] = None,
        headers: Optional[Dict[str, str]] = None,
        timeout: Optional[int] = None,
        file_transfer_timeout: Optional[int] = None,
        proxies: Optional[Dict[str, str]] = None,
        token: Optional[Union[Callable[[], str], str]] = None,
        token_url: Optional[str] = None,
        token_client_id: Optional[str] = None,
        token_client_secret: Optional[str] = None,
        token_scopes: Optional[List[str]] = None,
        token_custom_args: Optional[Dict[str, str]] = None,
        disable_pypi_version_check: Optional[bool] = None,
        debug: bool = False,
    ):
        super().__init__()

        self.api_key = api_key or self.api_key
        self.project = project or self.project
        self.client_name = client_name or self.client_name
        self.base_url = (base_url or self.base_url).rstrip("/")
        self.max_workers = max_workers or self.max_workers
        self.headers = headers or self.headers
        self.timeout = timeout or self.timeout
        self.file_transfer_timeout = file_transfer_timeout or self.file_transfer_timeout
        self.token = token
        self.proxies = proxies
        self.token_url = token_url or self.token_url
        self.token_client_id = token_client_id or self.token_client_id
        self.token_client_secret = token_client_secret or self.token_client_secret
        self.token_scopes = token_scopes or self.token_scopes
        self.token_custom_args = token_custom_args or self.token_custom_args
        self.api_subversion = api_subversion or self.api_subversion
        self.disable_pypi_version_check = (
            disable_pypi_version_check if disable_pypi_version_check is not None else self.disable_pypi_version_check
        )

        if self.api_key is None and self.token is None:
            self.token_custom_args.setdefault("verify", not self.disable_ssl)
            # If no api_key or token is present; try setting up a token generator
            token_generator = utils._token_generator.TokenGenerator(
                self.token_url,
                self.token_client_id,
                self.token_client_secret,
                self.token_scopes,
                self.token_custom_args,
            )

            if token_generator.token_params_set():
                self.token = token_generator.return_access_token

            if self.token is None:
                raise CogniteAPIKeyError("No API key or token or token generation arguments have been specified")

        if self.client_name is None:
            raise ValueError(
                "No client name has been specified. Pass it to the CogniteClient or set the environment variable "
                "'COGNITE_CLIENT_NAME'."
            )

        if debug:
            utils._logging._configure_logger_for_debug_mode()

        if not self.disable_pypi_version_check:
            try:
                utils._auxiliary._check_client_has_newest_major_version()
            except Exception:
                # PyPI is for some reason not reachable, skip version check
                pass

    def __str__(self):
        return pprint.pformat(self.__dict__, indent=4)

    def _repr_html_(self):
        return self.__str__()
